from threading import Event, RLock
from typing import Hashable, Union, Iterable, Collection, Optional, List

import numpy as np

from respice.collections import OneToManyInvertibleMapping
from respice.components import Component, Branch, SwitchBranch


class TransientSimulation:
    class _LockedOperationsInterface:
        def __init__(self, parent):
            self._parent = parent

        @property
        def ts(self):
            return self._parent._ts

        @property
        def node_potentials(self):
            return self._parent._node_potentials

        @property
        def branch_voltages(self):
            return self._parent._branch_voltages

        @property
        def branch_currents(self):
            return self._parent._branch_currents

        @property
        def component_states(self):
            return self._parent._component_states

        @property
        def switch_states(self):
            return self._parent._switch_states

        def extend_from_simulation(self, simulation):
            self.ts.extend(simulation.get_timesteps())
            for node in self._parent._nodes:
                self.node_potentials[node].extend(simulation.get_potentials(node))
            for branch in self._parent._components_to_branches.reverse:
                self.branch_voltages[branch].extend(simulation.get_voltages(branch))
                self.branch_currents[branch].extend(simulation.get_currents(branch))
                if isinstance(branch, SwitchBranch):
                    self.switch_states[branch].extend(simulation.get_switch_states(branch))
            for component in self._parent._components_to_branches:
                self.component_states[component].extend(simulation.get_states(component))

    def __init__(self,
                 ts1: float,
                 ts2: float,
                 components_to_branches: OneToManyInvertibleMapping,
                 nodes: Iterable[Hashable]):

        self._ts1 = ts1
        self._ts2 = ts2

        self._components_to_branches = OneToManyInvertibleMapping(components_to_branches)
        self._nodes = list(nodes)
        self._ts = []
        self._node_potentials = {node: [] for node in self._nodes}
        self._branch_voltages = {branch: [] for branch in components_to_branches.reverse}
        self._branch_currents = {branch: [] for branch in components_to_branches.reverse}
        self._component_states = {component: [] for component in components_to_branches}
        self._switch_states = {branch: []
                               for branch in components_to_branches.reverse
                               if isinstance(branch, SwitchBranch)}

        self._modification_lock = RLock()
        self._finished = Event()
        self._cancelled = Event()
        self._aborted = Event()
        self._exception = None

    def __enter__(self) -> _LockedOperationsInterface:
        """
        Acquire the *modification lock*.

        :return:
            A view into the object data allowing direct, thread-safe modifications.
        """
        self._modification_lock.acquire()
        if self._finished.is_set():
            self._modification_lock.release()
            raise RuntimeError('modifications after the simulation has finished are prohibited')
        return self._LockedOperationsInterface(self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Releases the modification lock.
        """
        self._modification_lock.release()

    def finish(self):
        """
        Signals the simulation to be finished.

        This function shouldn't be used from user perspective but is intended to be used from inside the simulation
        implementations.
        """
        with self._modification_lock:
            # Optimize inner fields for performance (e.g. convert lists to numpy arrays).
            self._ts = np.fromiter(self._ts, dtype=float)
            self._node_potentials = {node: np.fromiter(data, dtype=float) for node, data in self._node_potentials.items()}
            self._branch_voltages = {branch: np.fromiter(data, dtype=float) for branch, data in self._branch_voltages.items()}
            self._branch_currents = {branch: np.fromiter(data, dtype=float) for branch, data in self._branch_currents.items()}
            self._switch_states = {switch: np.fromiter(data, dtype=bool) for switch, data in self._switch_states.items()}

            self._finished.set()

    def cancel(self):
        """
        Cancels an ongoing simulation.

        If the simulation has finished or is about to be finished, this function has no effect except still
        setting `is_cancelled` to `True`.
        """
        self._cancelled.set()
        self._finished.set()

    @property
    def is_cancelled(self) -> bool:
        """
        Determines whether the simulation was cancelled.
        """
        return self._cancelled.is_set()

    def abort(self, ex: Exception):
        """
        Aborts the simulation due to an error.

        This function shouldn't be used from user perspective but is intended to be used from inside the simulation
        implementations.

        :param ex:
            Exception that occurred causing the simulatino to abort.
        """
        self._exception = ex
        self._aborted.set()
        self._finished.set()

    @property
    def is_aborted(self) -> bool:
        """
        Determines whether the simulation was aborted due to an error.
        """
        return self._aborted.is_set()

    def wait(self, timeout: Optional[float] = None, throw: bool = True) -> bool:
        """
        Wait for the simulation to finish.

        :param timeout:
            Timeout to wait. If `None`, blocks infinitely until result is ready.
        :param throw:
            Whether exceptions from the computation thread are thrown in the calling thread.
        :return:
            `True` if the result is ready, `False` if not. Only meaningful when a timeout is given.
        """
        finished = self._finished.wait(timeout)
        if throw and self._aborted.is_set():
            raise self._exception
        return finished

    def get_timesteps(self) -> np.ndarray:
        """
        Returns all discrete time steps simulated.
        """
        with self._modification_lock:
            return np.array(self._ts)

    t = get_timesteps  # Shorthand function

    def get_potentials(self, node: Hashable) -> np.ndarray:
        """
        Returns the calculated potentials at the given node.

        :param node:
            The node.
        """
        return np.array(self._node_potentials[node])

    def get_voltages(self, entity: Union[Component, Branch, Hashable], target_node: Hashable = None) -> np.ndarray:
        """
        Returns the solved voltages over a given entity.

        :param entity:
            Can be either:
            * a component (where `default_branch` must be set and not return `None` - e.g. two-terminal-components).
            * a branch.
            * a node.
        :param target_node:
            If `entity` is a (source) node, this parameter denotes the target node to get the voltages from in between.
        :return:
            The solved voltages measured in Volts in form of an `np.ndarray`.
        """
        if target_node is None:
            # Lookup by single component or branch.
            branch = entity.default_branch if isinstance(entity, Component) else entity

            with self._modification_lock:
                return np.array(self._branch_voltages[branch])
        else:
            # Lookup between nodes.
            with self._modification_lock:
                return self._node_potentials[target_node] - self._node_potentials[entity]

    v = get_voltages  # Shorthand function

    def get_currents(self, entity: Union[Component, Branch]) -> np.ndarray:
        """
        Returns the solved currents over a given entity.

        :param entity:
            Can be either:
            * a component (where `default_branch` must be set and not return `None` - e.g. two-terminal-components).
            * a branch.
        :return:
            The solved currents measured in Amperes in form of an `np.ndarray`.
        """
        branch = entity.default_branch if isinstance(entity, Component) else entity

        with self._modification_lock:
            return np.array(self._branch_currents[branch])

    i = get_currents  # Shorthand function

    def get_power(self, *entities: Union[Component, Branch]) -> np.ndarray:
        """
        Calculates power consumption over multiple components or branches.

        :param entities:
            (Multiple) entities to calculate power for. If a component is given, its default branch is taken.
        :return:
            The power measured in Watts.
        """
        with self._modification_lock:
            return sum(self.get_voltages(entity) * self.get_currents(entity)
                       for entity in entities)

    p = get_power  # Shorthand function

    def get_states(self, component: Component) -> List[np.ndarray]:
        """
        Returns the computed states of the component.

        :param component:
            The component.
        :return:
            The state vectors. Refer to each component type individually which entry means what state.
        """
        with self._modification_lock:
            return list(self._component_states[component])

    s = get_states  # Shorthand function

    def get_switch_states(self, switch: SwitchBranch) -> np.ndarray:
        """
        Returns the computed switch states during simulation.

        :param switch:
            The switch.
        :return:
            The switch states, an array with `bool`s.
        """
        with self._modification_lock:
            return np.array(self._switch_states[switch])

    def plot(self, *entities: Union[Component, Branch], features: Collection[str] = 'vip', throw: bool = True):
        """
        Plots the result.

        Plotting waits until the full simulation is finished. Live-plotting is right now not supported.

        Uses Plotly to present graphs in your browser. Make sure to have it installed.

        :param entities:
            Optional specific entities to plot. If none are specified, all available are plotted.
            Branches as well as components can be specified. In case the component defines a valid `default_branch`,
            this branch's data will be plotted. If `default_branch` returns `None`, all associated branches of this
            component will be plotted.

            Duplicate branches are filtered properly and won't be plotted twice.
        :param features:
            What kind of plots to make. Must be an iterable made out of following values:
            * v: voltage plot
            * i: current plot
            * p: power plot
            * s: component states
            * c: *c*ontrolled switch states
        :param throw:
            Whether exceptions from the computation thread are thrown in the calling thread.
            Set it to `False` if you also want to print partial results.
        """
        if not features:
            raise ValueError('no features specified')

        component_features = {}
        branch_features = {}
        for i, feature in enumerate(features, start=1):
            if feature == 'v' or feature == 'i' or feature == 'p' or feature == 'c':
                branch_features[i] = feature
            elif feature == 's':
                component_features[i] = feature
            else:
                raise ValueError(f'invalid feature: {feature}')

        self.wait(throw=throw)

        # Determine necessary components and respective branches to plot.
        if entities:
            plotset = {}
            for entity in entities:
                if isinstance(entity, Component):
                    component = entity
                    plotset[component] = (self._components_to_branches[component]
                                          if component.default_branch is None else
                                          {component.default_branch})
                else:
                    branch = entity
                    plotset[self._components_to_branches.reverse[branch]] = branch
        else:
            plotset = self._components_to_branches

        from itertools import cycle

        try:
            from plotly.subplots import make_subplots
            from plotly.graph_objects import Scatter
            from plotly.express import colors
        except ImportError:
            raise RuntimeError('plotly not found! Please install before using this function.')

        fig = make_subplots(rows=len(features), cols=1, shared_xaxes=True)

        colorcycle = cycle(colors.qualitative.Plotly)

        for component, branches in plotset.items():
            color = next(colorcycle)

            for i, feature in component_features.items():
                # Currently there is only the "state"-feature, no need for an if-else.
                for k, states in enumerate(zip(*self.get_states(component))):
                    fig.append_trace(
                        Scatter(
                            x=self._ts,
                            y=states,
                            name=f'state({component})[{k}]',
                            mode='lines',
                            line={'color': color},
                            showlegend=True,
                        ),
                        row=i,
                        col=1,
                    )

            for branch in branches:
                showlegend = True
                for i, feature in branch_features.items():
                    if feature == 'v':
                        get_solutions = self.get_voltages
                    elif feature == 'i':
                        get_solutions = self.get_currents
                    elif feature == 'p':
                        get_solutions = self.get_power
                    elif feature == 'c':
                        if not isinstance(branch, SwitchBranch):
                            continue

                        get_solutions = lambda branch: [1 if s else 0 for s in self.get_switch_states(branch)]
                    else:
                        raise AssertionError

                    fig.append_trace(
                        Scatter(
                            x=self._ts,
                            y=get_solutions(branch),
                            legendgroup=str(component),
                            name=str(branch),
                            mode='lines',
                            line={'color': color},
                            showlegend=showlegend,
                        ),
                        row=i,
                        col=1,
                    )

                    showlegend = False

        # Update xaxis properties
        axis_labels = {
            'v': 'voltage V [V]',
            'i': 'current I [A]',
            'p': 'power P [W]',
            's': 'state',
            'c': 'switch state',
        }

        for i, feature in enumerate(features, start=1):
            fig.update_yaxes(row=i, col=1, title_text=axis_labels[feature])

            if feature == 'c':
                fig.update_yaxes(
                    row=i,
                    col=1,
                    tickmode='array',
                    tickvals=[0, 1],
                    ticktext=['off', 'on'],
                )

        fig.update_xaxes(row=len(features), col=1, title_text="time t [s]")
        fig.update_layout(title="Transient Simulation Results")

        fig.show()

    def save(self, filename: str):
        """
        Stores quickly simulation data inside the given circuit to CSV-format.

        This functions waits until the simulation is finished completely.

        :param filename:
            The filename to store the data in.
        """
        self.wait()

        data = [('t', self._ts)]

        # Additionally iterating over components sorts branches by component,
        # thus makes it easier for humans to read.
        for component in self._components_to_branches:
            for branch in self._components_to_branches[component]:
                data.append((f'v({str(branch)})', self.get_voltages(branch)))
                data.append((f'i({str(branch)})', self.get_currents(branch)))
                if isinstance(branch, SwitchBranch):
                    data.append((f'switchstate({str(branch)})',
                                 ['on' if ss else 'off' for ss in self.get_switch_states(branch)]))

            for i, state in enumerate(zip(*self.get_states(component)), start=1):
                data.append((f'state({str(component)})[{i}]', state))

        import csv
        with open(filename, 'w', newline='') as fl:
            w = csv.writer(fl)
            w.writerow(name for name, _ in data)
            w.writerows(zip(*(d for _, d in data)))

    def print_progress(self):
        try:
            import rich.progress
            import pint
        except ImportError:
            raise RuntimeError('Required interactive package(s) not found. Please install before using this function.')

        progress = rich.progress.Progress(
            '[progress.description]{task.description}',
            rich.progress.BarColumn(),
            '[progress.percentage]{task.percentage:>3.1f}% ({task.fields[completed_str]}/{task.fields[total_str]})',
            '•',
            'ETC',
            rich.progress.TimeRemainingColumn(),
            refresh_per_second=2,
        )

        unitreg = pint.UnitRegistry()

        total = self._ts2 - self._ts1
        with progress:
            completed = self._ts[-1] - self._ts1 if len(self._ts) > 0 else 0.0
            task = progress.add_task(
                'simulate',
                total=total,
                completed=completed,
                total_str=f'{(total * unitreg.seconds).to_compact():.2f~P}',
                completed_str=f'{(completed * unitreg.seconds).to_compact():.2f~P}',
            )

            while not self._finished.wait(0.5):
                completed = self._ts[-1] - self._ts1 if len(self._ts) > 0 else 0.0
                progress.update(task, completed=completed, completed_str=f'{(completed * unitreg.seconds).to_compact():.2f~P}')

            progress.update(task, completed=total)



from collections import OrderedDict


# cachetools's Cache classes are not so fully-featured as they appear. They are simple and work for many things,
# but respice needs access to a few access statistics as well as hidden access without influencing the current cache
# order.
class LRUCache:
    """
    A cache implementation with LRU (least recently used) strategy.
    """
    def __init__(self, maxsize):
        self._data = OrderedDict()
        self._maxsize = maxsize

    @property
    def maxsize(self):
        """
        Maximum size of the cache.
        """
        return self._maxsize

    @maxsize.setter
    def maxsize(self, value):
        while value < len(self._data):
            self._data.popitem()

        self._maxsize = value

    def __len__(self):
        """
        Returns how many elements are inside the cache.
        """
        return len(self._data)

    @property
    def full(self):
        """
        Returns `True` if the cache is filled up (`len(cache) == cache.maxsize`), else `False`.
        """
        return len(self._data) >= self._maxsize

    def __getitem__(self, key):
        """
        Returns a cache item.

        :param key:
            Key of cache item to retrieve.
        """
        self.update(key)
        return self._data[key]

    def get_stealth(self, key):
        """
        Returns a cache entry. This is a stealth operation, i.e. it returns the cache item without resetting
        its associated LRU value.

        :param key:
            Key of cache item to retrieve.
        """
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value
        self.update(key)
        if len(self._data) > self.maxsize:
            self._data.popitem()

    def set_stealth(self, key, value):
        """
        Sets a cache entry as a stealth operation, i.e. add the key considering it is the least up-to-date one.
        If the cache is already full, this is a no-op.
        """
        if not self.full:
            self._data[key] = value

    def __delitem__(self, key):
        """
        Deletes a cache entry.
        """
        del self._data[key]

    def update(self, key):
        """
        Resets the LRU value of a cache entry.
        """
        self._data.move_to_end(key, last=False)

    def __iter__(self):
        """
        Iterator over all cache entry keys.
        """
        return iter(self._data.keys())

    def last(self):
        """
        Returns the last used cache entry.
        """
        try:
            return next(iter(self._data.items()))
        except StopIteration:
            raise KeyError('cache is empty') from None

    def __contains__(self, key):
        return key in self._data

    def keys(self):
        """
        Retrieves all keys inside the cache in order of last usage.

        This is a stealth operation, i.e. iterating over the cache entries **will not** change their order in the cache.
        To do so, call `update(key)` appropriately. However note, that this cannot be done from within iterating, as
        it changes the underlying data structure. To still do so raises an error.
        """
        return self.__iter__()

    def values(self):
        """
        Retrieves all values inside the cache in order of last usage.

        This is a stealth operation, i.e. iterating over the cache entries **will not** change their order in the cache.
        To do so, call `update(key)` appropriately. However note, that this cannot be done from within iterating, as
        it changes the underlying data structure. To still do so raises an error.
        """
        return iter(self._data.values())

    def items(self):
        """
        Retrieves all items (key-value pairs) inside the cache in order of last usage.

        This is a stealth operation, i.e. iterating over the cache entries **will not** change their order in the cache.
        To do so, call `update(key)` appropriately. However note, that this cannot be done from within iterating, as
        it changes the underlying data structure. To still do so raises an error.
        """
        return iter(self._data.items())

    def clear(self):
        """
        Empties the cache.
        """
        self._data.clear()

    def __repr__(self):
        return f'<{type(self).__name__} at {hex(id(self))} (capacity: {len(self._data)}/{self._maxsize} used)>'



from dataclasses import dataclass, field
from typing import Callable

import numpy as np

from respice.math.numerics import trap, trap_jac
from .TwoTerminalCurrentComponent import TwoTerminalCurrentComponent


@dataclass(eq=False)
class Motor(TwoTerminalCurrentComponent):
    r"""
    Describes a motor model according to following model:

    .. math::

        K_T i - T_L &= J \dot \omega + D \omega \\
        v &= i R + \dot i L + K_E \omega

    where :math:`i` denotes input current, :math:`v` terminal voltage and :math:`\omega` angular speed.

    The model supports an additional load torque :math:`T_L`, which can be an arbitrary function of time.

    R:
        Terminal resistance measured in Ohm (:math:`\Omega`).
    L:
        Internal inductance measured in Henry (:math:`H`).
    KT:
        Torque constant measured in :math:`Nm / A`.
    KE:
        Back-EMF constant measured in :math:`V \cdot rad / s`.
    J:
        Moment of inertia of rotor measured in :math:`kg \cdot m^2`.
    D:
        Damping constant due to viscous friction in :math:`Nm`.
    TL:
        A time-(only)-variate function specifying a load-torque depending on time. By default 0.
    state_voltage:
        The initial voltage of the motor. By default 0.
    state_current:
        The initial current of the motor. By default 0.
    state_angular_frequency:
        The initial angular frequency of the motor axis. By default 0.
    """

    R: float
    L: float
    KT: float
    KE: float
    J: float
    D: float
    TL: Callable[[float], float] = lambda t: 0
    state_voltage: float = field(default=0.0, repr=False)
    state_current: float = field(default=0.0, repr=False)
    state_angular_frequency: float = field(default=0.0, repr=False)

    def _get_state_space_model(self, t: float, v: float):
        A = np.array([
            [-self.R / self.L, -self.KE / self.L],
            [self.KT / self.J, -self.D / self.J],
        ])

        b = np.array([
            v / self.L,
            -self.TL(t) / self.J,
        ])

        return A, b

    def _get_state_space_model_jacobian(self, t: float, v: float):
        dA = np.zeros((1, 2, 2))
        db = np.array([
            [1 / self.L],
            [0],
        ])
        return dA, db

    def _calculate_next_state(self, v: float, t1: float, t2: float):
        x1 = np.array([self.state_current, self.state_angular_frequency])

        A1, b1 = self._get_state_space_model(t1, self.state_voltage)
        A2, b2 = self._get_state_space_model(t2, v)

        return trap(x1, t2 - t1, A1, b1, A2, b2)

    def get_current(self, v: float, t1: float, t2: float) -> float:
        i, _ = self._calculate_next_state(v, t1, t2)
        return i

    def get_jacobian(self, v: float, t1: float, t2: float) -> float:
        A2, b2 = self._get_state_space_model(t2, v)
        dA, db = self._get_state_space_model_jacobian(t2, v)

        return trap_jac(
            self._calculate_next_state(v, t1, t2),
            t2 - t1,
            A2,
            dA,
            db,
        )[0, 0]

    def update(self, v_i: np.ndarray, t1: float, t2: float):
        v = v_i[0]
        i, omega = self._calculate_next_state(v, t1, t2)

        self.state_current = i
        self.state_voltage = v
        self.state_angular_frequency = omega

    @property
    def state(self):
        return self.state_current, self.state_voltage, self.state_angular_frequency

    @state.setter
    def state(self, val):
        self.state_current, self.state_voltage, self.state_angular_frequency = val



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Hashable, List, Tuple

from .Branch import Branch, SwitchBranch
from .Component import Component


@dataclass(eq=False)
class Switch(Component):
    """
    The base class for all simple switches.
    """
    _switch: _SwitchBranch = field(init=False, repr=False)

    @dataclass(eq=False)
    class _SwitchBranch(SwitchBranch):
        """
        The branch that represents the simple switch.

        Effectively mirrors all functions like `get_switch_state` back to the parent component.
        """
        def switch_state(self, t: float) -> bool:
            return self.component.switch_state(t)

        def __str__(self):
            return f'{str(self.component)}'

    def __post_init__(self):
        self._switch = self._SwitchBranch(self)

    def connect(self, terminal1: Hashable, terminal2: Hashable) -> List[Tuple[Hashable, Hashable, Branch]]:
        return [(terminal1, terminal2, self._switch)]

    def switch_state(self, t: float) -> bool:
        raise NotImplementedError

    @property
    def default_branch(self):
        """
        Returns the one and only branch of a two terminal component.
        """
        return self._switch



from itertools import chain, islice, tee


def flatten(list_of_lists):
    """
    Flatten one level of nesting.

    From https://docs.python.org/3.7/library/itertools.html#itertools-recipes.
    """
    return chain.from_iterable(list_of_lists)


def pairwise(iterable):
    """
    s -> (s0,s1), (s1,s2), (s2, s3)

    From https://docs.python.org/3.7/library/itertools.html#itertools-recipes.
    """
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def take(n, iterable):
    """
    Return first n items of the iterable as a list.

    From https://docs.python.org/3/library/itertools.html#itertools-recipes.
    """
    return list(islice(iterable, n))



"""
Implementation of the "overshoot" type solver, an enhanced wrapper around SciPy's `root`.

It is a naive and simple solver which attempts to overshoot solutions on stationary locations and retries solving.
"""
import math
import warnings
from typing import Callable

import numpy as np
from scipy.optimize import root, OptimizeResult


# This solver is deprecated! It exists solely for legacy and compatibility purposes!


class UnmetPrecisionWarning(UserWarning):
    def __init__(self, msg: str, result: OptimizeResult):
        super().__init__(msg)

        self.result = result


def _disturb_solution(trial: int, solution: OptimizeResult, factor_multiplier=100) -> np.ndarray:
    """
    Disturbs the result of a solution. This function is used when retrying the solution because convergence
    didn't succeed and has gotten so worse that the solver stops further trying. Trying with a different starting
    point is the recommended method.

    This function specifically applies relative disturbance and considers the direction of the Jacobian matrix
    to effectively over-shoot the current (non-converged) solution. The higher the trial number, the higher
    the overshooting.

    :param trial:
        The current trial number (starts from 1).
    :param solution:
        The `OptimizationResult` as received from `scipy.optimize.root`.
    :param factor_multiplier:
        The disturbance factor to the trial.
    :return:
        The disturbed solution vector.
    """
    directional_multiplier = np.array([(-1 if v < 0 else 1) for v in solution.fjac @ solution.x])
    return directional_multiplier * factor_multiplier ** trial * solution.x


def solve(eq: Callable, x0: np.ndarray, args=tuple(), jac=None, method='hybr') -> OptimizeResult:
    """
    Attempts to solve the given electrical-circuit-governing system equation.

    Additionally employs retries if convergence tolerances are not reached with different starting points
    utilizing `_disturb_solution`. If `_disturb_solution` does yield conversion to the same (or rather very close
    point), then again a retry is attempted, but with a point that's more far away. If other points are hit that
    still don't fulfill convergence restrictions, retries are attempted on them as well. Retries are tracked per
    solution, so in case that we encounter cycles where we jump to different but non-satisfying solutions and come
    back to a solution we already had, `_disturb_solution` will attempt a different start vector never used before
    to not end up in useless wastes of trials.

    By default int((math.log10(len(start_vector)) + 1) * 4) trials will be attempted if needed, so for larger
    systems more points are investigated for bad convergence. To not increase evaluation time too much, the
    logarithm is used to only scale to the order of the system.

    :param eq:
        The system equation function to solve.
    :param x0:
        The initial start vector.
    :param args:
        Args to `eq` and `jac`.
    :param jac:
        The Jacobian (function) of eq.
    :param method:
        Which root solver method to use.
        See https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.root.html.
    :return:
        The solution (or on non-convergence, the closest solution found during the process).
    """
    unsatisfying_results = []
    trials = {}
    retries = int((math.log10(len(x0)) + 1) * 4)
    for trial in range(retries):
        result = root(
            eq,
            x0,
            args=args,
            method=method,
            jac=jac,
        )

        if result.success:
            break

        unsatisfying_results.append(result)

        # Result disturbance is a bit smarter and resets the trial count if we really found a different point,
        # so in case the next different value again does not converge, we do not overshoot extremely just because
        # this happened at a high trial count. However, if we might re-encounter the same solution, then the old
        # trial count gets restored again. That ensures that we really always try out new values and aren't driven
        # into useless re-evaluation of the same points until all trials are exhausted.
        for solution in trials:
            if math.isclose(np.linalg.norm(solution), np.linalg.norm(result.x)):
                trials[solution] += 1
                result_trials = trials[solution]
                break
        else:
            trials[tuple(result.x)] = 1  # numpy arrays are not hashable, so we produce a hashable vector.
            result_trials = 1

        x0 = _disturb_solution(result_trials, result)
    else:
        # Filter out best result closest to zero.
        result = min(unsatisfying_results, key=lambda x: np.linalg.norm(x.fun))

        warnings.warn(UnmetPrecisionWarning(
            'Failed to converge after several retries. Taking the closest result as granted.',
            result,
        ))

    return result



from itertools import chain
from threading import RLock, Event
from typing import Iterable, List, Hashable, Tuple, Union, Collection, Optional

import numpy as np

from respice.collections import OneToManyInvertibleMapping
from respice.components import Component, SwitchBranch, Branch
from respice.itertools import flatten, pairwise
from respice.itertools.custom import deresolve, intersperse
from .TransientSimulation import TransientSimulation


class TransientEFMSimulation:
    class _LockedOperationsInterface:
        def __init__(self, parent):
            self._parent = parent

        @property
        def tss(self):
            return self._parent._tss

        @property
        def node_potentials(self):
            return self._parent._node_potentials

        @property
        def branch_voltages(self):
            return self._parent._branch_voltages

        @property
        def branch_currents(self):
            return self._parent._branch_currents

        @property
        def component_states(self):
            return self._parent._component_states

        @property
        def switch_states(self):
            return self._parent._switch_states

        def append_from_simulation(self, simulation: TransientSimulation):
            self.tss.append(simulation.get_timesteps())
            for node in self._parent._nodes:
                self.node_potentials[node].append(simulation.get_potentials(node))
            for branch in self._parent._components_to_branches.reverse:
                self.branch_voltages[branch].append(simulation.get_voltages(branch))
                self.branch_currents[branch].append(simulation.get_currents(branch))
                if isinstance(branch, SwitchBranch):
                    self.switch_states[branch].append(simulation.get_switch_states(branch))
            for component in self._parent._components_to_branches:
                self.component_states[component].append(simulation.get_states(component))

        def extend_from_simulation(self, simulation):
            self.tss.extend(simulation.get_interval_timesteps())
            for node in self._parent._nodes:
                self.node_potentials[node].extend(simulation.get_interval_potentials(node))
            for branch in self._parent._components_to_branches.reverse:
                self.branch_voltages[branch].extend(simulation.get_interval_voltages(branch))
                self.branch_currents[branch].extend(simulation.get_interval_currents(branch))
                if isinstance(branch, SwitchBranch):
                    self.switch_states[branch].extend(simulation.get_interval_switch_states(branch))
            for component in self._parent._components_to_branches:
                self.component_states[component].extend(simulation.get_interval_states(component))

    def __init__(self,
                 skips: List[float],
                 components_to_branches: OneToManyInvertibleMapping,
                 nodes: Iterable[Hashable]):

        self._skips = list(skips)

        self._components_to_branches = OneToManyInvertibleMapping(components_to_branches)
        self._nodes = list(nodes)
        self._tss = []
        self._node_potentials = {node: [] for node in self._nodes}
        self._branch_voltages = {branch: [] for branch in components_to_branches.reverse}
        self._branch_currents = {branch: [] for branch in components_to_branches.reverse}
        self._component_states = {component: [] for component in components_to_branches}
        self._switch_states = {branch: []
                               for branch in components_to_branches.reverse
                               if isinstance(branch, SwitchBranch)}

        self._modification_lock = RLock()
        self._finished = Event()
        self._cancelled = Event()
        self._aborted = Event()
        self._exception = None

    def __enter__(self) -> _LockedOperationsInterface:
        """
        Acquire the *modification lock*.

        :return:
            A view into the object data allowing direct, thread-safe modifications.
        """
        self._modification_lock.acquire()
        if self._finished.is_set():
            self._modification_lock.release()
            raise RuntimeError('modifications after the simulation has finished are prohibited')
        return self._LockedOperationsInterface(self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Releases the modification lock.
        """
        self._modification_lock.release()

    def finish(self):
        """
        Signals the simulation to be finished.

        This function shouldn't be used from user perspective but is intended to be used from inside the simulation
        implementations.
        """
        with self._modification_lock:
            self._finished.set()

    def cancel(self):
        """
        Cancels an ongoing simulation.

        If the simulation has finished or is about to be finished, this function has no effect except still
        setting `is_cancelled` to `True`.
        """
        self._cancelled.set()
        self._finished.set()

    @property
    def is_cancelled(self) -> bool:
        """
        Determines whether the simulation was cancelled.
        """
        return self._cancelled.is_set()

    def abort(self, ex: Exception):
        """
        Aborts the simulation due to an error.

        This function shouldn't be used from user perspective but is intended to be used from inside the simulation
        implementations.

        :param ex:
            Exception that occurred causing the simulatino to abort.
        """
        self._exception = ex
        self._aborted.set()
        self._finished.set()

    @property
    def is_aborted(self) -> bool:
        """
        Determines whether the simulation was aborted due to an error.
        """
        return self._aborted.is_set()

    def wait(self, timeout: Optional[float] = None, throw: bool = True) -> bool:
        """
        Wait for the simulation to finish.

        :param timeout:
            Timeout to wait. If `None`, blocks infinitely until result is ready.
        :param throw:
            Whether exceptions from the computation thread are thrown in the calling thread.
        :return:
            `True` if the result is ready, `False` if not. Only meaningful when a timeout is given.
        """
        finished = self._finished.wait(timeout)
        if throw and self._aborted.is_set():
            raise self._exception
        return finished

    @staticmethod
    def _interpolate_intervals(tss, intervals, skips):
        interpolated_ts = list(tss[0])
        interpolated_vs = list(intervals[0])

        for ss, ((ts1, vs1), (ts2, vs2)) in zip(skips, pairwise(zip(tss, intervals))):
            merged_relative_ts = np.fromiter(
                deresolve(sorted(chain(ts1 - ts1[0], ts2 - ts2[0]))),
                dtype=float)

            distributed_ts = np.fromiter(
                flatten(merged_relative_ts + t
                        for t in np.linspace(ts1[-1], ts2[0], ss, endpoint=False)),
                dtype=float)

            interpolated_ts.extend(distributed_ts)

            linear_ramp = (distributed_ts - ts1[-1]) / (ts2[0] - ts1[-1])

            # Normalized voltages with the slope of the envelope taken into account to avoid jumps in the interpolation.
            nvs1 = vs1 - np.interp(ts1, [ts1[0], ts1[-1]], [vs1[0] - vs1[-1], 0])
            nvs2 = vs2 - np.interp(ts2, [ts2[0], ts2[-1]], [0, vs2[-1] - vs2[0]])

            interpolated_vs.extend(
                np.tile(np.interp(merged_relative_ts, ts1 - ts1[0], nvs1), ss) * (1 - linear_ramp) +
                np.tile(np.interp(merged_relative_ts, ts2 - ts2[0], nvs2), ss) * linear_ramp
            )

            # Also append the real result to interpolation.
            interpolated_ts.extend(ts2)
            interpolated_vs.extend(vs2)

        return np.fromiter(interpolated_ts, dtype=float), np.fromiter(interpolated_vs, dtype=float)

    @staticmethod
    def _retrieve_envelope(intervals):
        return TransientEFMSimulation._straighten([x[0], x[-1]] for x in intervals)

    @staticmethod
    def _straighten(intervals):
        return np.fromiter(flatten(intervals), dtype=float)

    def get_interval_timesteps(self) -> List[np.ndarray]:
        """
        Returns all discrete time steps simulated, structured accordingly into EFM intervals.
        """
        with self._modification_lock:
            return list(np.array(ts) for ts in self._tss)

    def get_timesteps(self) -> np.ndarray:
        """
        Returns all discrete time steps simulated.
        """
        with self._modification_lock:
            return self._straighten(self._tss)

    def get_envelope_timesteps(self) -> np.ndarray:
        """
        Returns all discrete time steps of the simulation envelope.
        """
        with self._modification_lock:
            return self._retrieve_envelope(self._tss)

    def get_interval_potentials(self, node: Hashable) -> List[np.ndarray]:
        """
        Returns the calculated potentials (in Volts) at the given node, structured accordingly into EFM intervals.

        :param node:
            The node.
        """
        with self._modification_lock:
            return self._node_potentials[node]

    def get_potentials(self, node: Hashable) -> np.ndarray:
        """
        Returns the calculated potentials (in Volts) at the given node.

        :param node:
            The node.
        """
        return self._straighten(self.get_interval_potentials(node))

    def get_envelope_potentials(self, node: Hashable) -> np.ndarray:
        """
        Returns the calculated potentials (in Volts) at the given node of the simulation envelope.

        :param node:
            The node.
        """
        return self._retrieve_envelope(self.get_interval_potentials(node))

    def get_interpolated_potentials(self, node: Hashable) -> Tuple[np.ndarray, np.ndarray]:
        r"""
        Creates an interpolation using the solved potentials.

        The EFM (Envelope Following Method) skips over large intervals. For these skipped intervals, this function
        creates a simple interpolation :math:`f_{m \leftrightarrow n}` by linearly blending two adjacent EFM intervals
        :math:`f_m` and :math:`f_n` with the blend function :math:`q` repeating the interval across the skipped section.

        .. math::

            (m + 1) T < t < n T: & q(t) = \frac{t - (m+1)T}{(n - m - 1)T}, \\
            & f_{m \leftrightarrow n}(t) = (1 - q) \cdot f_m(t \mod T) + q \cdot f_n(t \mod T)

        :param node:
            The node.
        :return:
            The interpolation time steps measured in seconds and interpolated potentials measured in Volts.
        """
        with self._modification_lock:
            return self._interpolate_intervals(
                self.get_interval_timesteps(),
                self.get_interval_potentials(node),
                self._skips,
            )

    t = get_timesteps  # Shorthand function

    def get_interval_voltages(self,
                              entity: Union[Component, Branch, Hashable],
                              target_node: Hashable = None) -> List[np.ndarray]:
        """
        Returns the solved voltages over a given entity, structured accordingly into EFM intervals.

        :param entity:
            Can be either:
            * a component (where `default_branch` must be set and not return `None` - e.g. two-terminal-components).
            * a branch.
            * a node.
        :param target_node:
            If `entity` is a (source) node, this parameter denotes the target node to get the voltages from in between.
        :return:
            The solved EFM-interval voltages measured in Volts as a list of `np.ndarray`s.
        """
        if target_node is None:
            # Lookup by single component or branch.
            branch = entity.default_branch if isinstance(entity, Component) else entity

            with self._modification_lock:
                return [np.array(voltages) for voltages in self._branch_voltages[branch]]
        else:
            # Lookup between nodes.
            with self._modification_lock:
                return [np.fromiter((target - source for target, source in zip(targets, sources)),
                                    dtype=float)
                        for targets, sources in zip(self._node_potentials[target_node],
                                                    self._node_potentials[entity])]

    def get_voltages(self, entity: Union[Component, Branch, Hashable], target_node: Hashable = None) -> np.ndarray:
        """
        Returns the solved voltages over a given entity.

        :param entity:
            Can be either:
            * a component (where `default_branch` must be set and not return `None` - e.g. two-terminal-components).
            * a branch.
            * a node.
        :param target_node:
            If `entity` is a (source) node, this parameter denotes the target node to get the voltages from in between.
        :return:
            The solved voltages measured in Volts in form of an `np.ndarray`.
        """
        return self._straighten(self.get_interval_voltages(entity, target_node))

    v = get_voltages  # Shorthand function

    def get_envelope_voltages(self,
                              entity: Union[Component, Branch, Hashable],
                              target_node: Hashable = None) -> np.ndarray:
        """
        Returns the solved voltages over a given entity of the simulation envelope.

        :param entity:
            Can be either:
            * a component (where `default_branch` must be set and not return `None` - e.g. two-terminal-components).
            * a branch.
            * a node.
        :param target_node:
            If `entity` is a (source) node, this parameter denotes the target node to get the voltages from in between.
        :return:
            The solved voltages measured in Volts in form of an `np.ndarray`.
        """
        return self._retrieve_envelope(self.get_interval_voltages(entity, target_node))

    def get_interpolated_voltages(self,
                                  entity: Union[Component, Branch, Hashable],
                                  target_node: Hashable = None) -> Tuple[np.ndarray, np.ndarray]:
        r"""
        Creates an interpolation using the solved voltages.

        The EFM (Envelope Following Method) skips over large intervals. For these skipped intervals, this function
        creates a simple interpolation :math:`f_{m \leftrightarrow n}` by linearly blending two adjacent EFM intervals
        :math:`f_m` and :math:`f_n` with the blend function :math:`q` repeating the interval across the skipped section.

        .. math::

            (m + 1) T < t < n T: & q(t) = \frac{t - (m+1)T}{(n - m - 1)T}, \\
            & f_{m \leftrightarrow n}(t) = (1 - q) \cdot f_m(t \mod T) + q \cdot f_n(t \mod T)

        :param entity:
            Can be either:
            * a component (where `default_branch` must be set and not return `None` - e.g. two-terminal-components).
            * a branch.
            * a node.
        :param target_node:
            If `entity` is a (source) node, this parameter denotes the target node to get the voltages from in between.
        :return:
            The interpolation time steps measured in seconds and interpolated voltages measured in Volts.
        """
        with self._modification_lock:
            return self._interpolate_intervals(
                self.get_interval_timesteps(),
                self.get_interval_voltages(entity, target_node),
                self._skips,
            )

    def get_interval_currents(self, entity: Union[Component, Branch]) -> List[np.ndarray]:
        """
        Returns the solved currents over a given entity, structured accordingly into EFM intervals.

        :param entity:
            Can be either:
            * a component (where `default_branch` must be set and not return `None` - e.g. two-terminal-components).
            * a branch.
        :return:
            The solved currents measured in Amperes in form of an `np.ndarray`.
        """
        branch = entity.default_branch if isinstance(entity, Component) else entity

        with self._modification_lock:
            return [np.array(currents) for currents in self._branch_currents[branch]]

    def get_currents(self, entity: Union[Component, Branch]) -> np.ndarray:
        """
        Returns the solved currents over a given entity.

        :param entity:
            Can be either:
            * a component (where `default_branch` must be set and not return `None` - e.g. two-terminal-components).
            * a branch.
        :return:
            The solved currents measured in Amperes in form of an `np.ndarray`.
        """
        return self._straighten(self.get_interval_currents(entity))

    i = get_currents  # Shorthand function

    def get_envelope_currents(self, entity: Union[Component, Branch]) -> np.ndarray:
        """
        Returns the solved currents over a given entity of the simulation envelope.

        :param entity:
            Can be either:
            * a component (where `default_branch` must be set and not return `None` - e.g. two-terminal-components).
            * a branch.
        :return:
            The solved currents measured in Amperes in form of an `np.ndarray`.
        """
        return self._retrieve_envelope(self.get_interval_currents(entity))

    def get_interpolated_currents(self, entity: Union[Component, Branch]) -> Tuple[np.ndarray, np.ndarray]:
        r"""
        Creates an interpolation using the solved currents.

        The EFM (Envelope Following Method) skips over large intervals. For these skipped intervals, this function
        creates a simple interpolation :math:`f_{m \leftrightarrow n}` by linearly blending two adjacent EFM intervals
        :math:`f_m` and :math:`f_n` with the blend function :math:`q` repeating the interval across the skipped section.

        .. math::

            (m + 1) T < t < n T: & q(t) = \frac{t - (m+1)T}{(n - m - 1)T}, \\
            & f_{m \leftrightarrow n}(t) = (1 - q) \cdot f_m(t \mod T) + q \cdot f_n(t \mod T)

        :param entity:
            Can be either:
            * a component (where `default_branch` must be set and not return `None` - e.g. two-terminal-components).
            * a branch.
        :return:
            The interpolation time steps measured in seconds and interpolated currents measured in Amperes.
        """
        with self._modification_lock:
            return self._interpolate_intervals(
                self.get_interval_timesteps(),
                self.get_interval_currents(entity),
                self._skips,
            )

    def get_interval_power(self, *entities: Union[Component, Branch]) -> List[np.ndarray]:
        """
        Calculates power consumption over multiple components or branches, structured accordingly into EFM intervals.

        :param entities:
            (Multiple) entities to calculate power for. If a component is given, its default branch is taken.
        :return:
            The power measured in Watts.
        """
        with self._modification_lock:
            powers = ([[v * i for v, i in zip(intv_v, intv_i)]
                       for intv_v, intv_i in zip(self.get_interval_voltages(entity),
                                                 self.get_interval_currents(entity))]
                      for entity in entities)

            return [np.fromiter((sum(p) for p in zip(*intv_p)), dtype=float) for intv_p in zip(*powers)]

    def get_power(self, *entities: Union[Component, Branch]) -> np.ndarray:
        """
        Calculates power consumption over multiple components or branches.

        :param entities:
            (Multiple) entities to calculate power for. If a component is given, its default branch is taken.
        :return:
            The power measured in Watts.
        """
        return self._straighten(self.get_interval_power(*entities))

    def get_envelope_power(self, *entities: Union[Component, Branch]) -> np.ndarray:
        """
        Calculates power consumption over multiple components or branches of the simulation envelope.

        :param entities:
            (Multiple) entities to calculate power for. If a component is given, its default branch is taken.
        :return:
            The power measured in Watts.
        """
        return self._retrieve_envelope(self.get_interval_power(*entities))

    def get_interpolated_power(self, *entities: Union[Component, Branch]) -> Tuple[np.ndarray, np.ndarray]:
        r"""
        Creates an interpolation using the solved powers.

        The EFM (Envelope Following Method) skips over large intervals. For these skipped intervals, this function
        creates a simple interpolation :math:`f_{m \leftrightarrow n}` by linearly blending two adjacent EFM intervals
        :math:`f_m` and :math:`f_n` with the blend function :math:`q` repeating the interval across the skipped section.

        .. math::

            (m + 1) T < t < n T: & q(t) = \frac{t - (m+1)T}{(n - m - 1)T}, \\
            & f_{m \leftrightarrow n}(t) = (1 - q) \cdot f_m(t \mod T) + q \cdot f_n(t \mod T)

        :param entities:
            (Multiple) entities to calculate power for. If a component is given, its default branch is taken.
        :return:
            The interpolation time steps measured in seconds and interpolated powers measured in Watts.
        """
        with self._modification_lock:
            return self._interpolate_intervals(
                self.get_interval_timesteps(),
                self.get_interval_power(*entities),
                self._skips,
            )

    p = get_power  # Shorthand function

    def get_interval_states(self, component: Component) -> List[List[np.ndarray]]:
        """
        Returns the computed states of the component, structured accordingly to EFM intervals.

        :param component:
            The component.
        :return:
            The state vectors. Refer to each component type individually which entry means what state.
        """
        with self._modification_lock:
            return [[np.array(state) for state in states] for states in self._component_states[component]]

    def get_states(self, component: Component) -> List[np.ndarray]:
        """
        Returns the computed states of the component.

        :param component:
            The component.
        :return:
            The state vectors. Refer to each component type individually which entry means what state.
        """
        return list(flatten(self.get_interval_states(component)))

    def get_envelope_states(self, component: Component) -> List[np.ndarray]:
        """
        Returns the computed states of the component of the simulation envelope.

        :param component:
            The component.
        :return:
            The state vectors. Refer to each component type individually which entry means what state.
        """
        return list(flatten([x[0], x[-1]] for x in self.get_interval_states(component)))

    def get_interpolated_states(self, component: Component) -> Tuple[np.ndarray, np.ndarray]:
        r"""
        Creates an interpolation using the solved component states.

        The EFM (Envelope Following Method) skips over large intervals. For these skipped intervals, this function
        creates a simple interpolation :math:`f_{m \leftrightarrow n}` by linearly blending two adjacent EFM intervals
        :math:`f_m` and :math:`f_n` with the blend function :math:`q` repeating the interval across the skipped section.

        .. math::

            (m + 1) T < t < n T: & q(t) = \frac{t - (m+1)T}{(n - m - 1)T}, \\
            & f_{m \leftrightarrow n}(t) = (1 - q) \cdot f_m(t \mod T) + q \cdot f_n(t \mod T)

        :param component:
            The component.
        :return:
            The interpolation time steps measured in seconds and interpolated states.
        """
        with self._modification_lock:
            ts = np.zeros(0)
            interpolations = []
            for states in zip(*(list(zip(*interval)) for interval in self.get_interval_states(component))):
                ts, interpolated = self._interpolate_intervals(self.get_interval_timesteps(), states, self._skips)
                interpolations.append(interpolated)

            return ts, np.transpose(interpolations)

    s = get_states  # Shorthand function

    def get_interval_switch_states(self, switch: SwitchBranch) -> List[np.ndarray]:
        """
        Returns the computed switch states during simulation, structured accordingly into EFM intervals.

        :param switch:
            The switch.
        :return:
            The switch states, an array with `bool`s.
        """
        with self._modification_lock:
            return [np.array(states) for states in self._switch_states[switch]]

    def get_switch_states(self, switch: SwitchBranch) -> np.ndarray:
        """
        Returns the computed switch states during simulation.

        :param switch:
            The switch.
        :return:
            The switch states, an array with `bool`s.
        """
        return np.fromiter(flatten(self.get_interval_switch_states(switch)), dtype=bool)

    def plot(self, *entities: Union[Component, Branch], features: Collection[str] = 'vip', throw: bool = True):
        """
        Plots the result.

        Uses Plotly to present graphs in your browser. Make sure to have it installed.

        :param entities:
            Optional specific entities to plot. If none are specified, all available are plotted.
            Branches as well as components can be specified. In case the component defines a valid `default_branch`,
            this branch's data will be plotted. If `default_branch` returns `None`, all associated branches of this
            component will be plotted.

            Duplicate branches are filtered properly and won't be plotted twice.
        :param features:
            What kind of plots to make. Must be an iterable made out of following values:
            * v: voltage plot
            * i: current plot
            * p: power plot
            * s: component states
            * c: *c*ontrolled switch states
        :param throw:
            Whether exceptions from the computation thread are thrown in the calling thread.
            Set it to `False` if you also want to print partial results.
        """
        if not features:
            raise ValueError('no features specified')

        component_features = {}
        branch_features = {}
        for i, feature in enumerate(features, start=1):
            if feature == 'v' or feature == 'i' or feature == 'p' or feature == 'c':
                branch_features[i] = feature
            elif feature == 's':
                component_features[i] = feature
            else:
                raise ValueError(f'invalid feature: {feature}')

        self.wait(throw=throw)

        # Determine necessary components and respective branches to plot.
        if entities:
            plotset = {}
            for entity in entities:
                if isinstance(entity, Component):
                    component = entity
                    plotset[component] = (self._components_to_branches[component]
                                          if component.default_branch is None else
                                          {component.default_branch})
                else:
                    branch = entity
                    plotset[self._components_to_branches.reverse[branch]] = branch
        else:
            plotset = self._components_to_branches

        from itertools import cycle

        try:
            from plotly.subplots import make_subplots
            from plotly.graph_objects import Scatter
            from plotly.express import colors
        except ImportError:
            raise RuntimeError('plotly not found! Please install before using this function.')

        fig = make_subplots(rows=len(features), cols=1, shared_xaxes=True)

        colorcycle = cycle(colors.qualitative.Plotly)

        for component, branches in plotset.items():
            color = next(colorcycle)

            for i, feature in component_features.items():
                # Currently there is only the "state"-feature, no need for an if-else.

                intervals = list(zip(*(list(zip(*interval)) for interval in self.get_interval_states(component))))
                envelopes = list(zip(*self.get_envelope_states(component)))
                ts, interpolations = self.get_interpolated_states(component)
                interpolations = np.transpose(interpolations)

                for k in range(len(intervals)):
                    fig.append_trace(
                        Scatter(
                            x=list(intersperse(self._tss, None)),
                            y=list(intersperse(intervals[k], None)),
                            legendgroup=f'state({str(component)})[{k}] (simulated)',
                            name=f'state({component})[{k}] (simulated)',
                            mode='lines',
                            line={'color': color},
                            showlegend=True,
                        ),
                        row=i,
                        col=1,
                    )

                    fig.append_trace(
                        Scatter(
                            x=self.get_envelope_timesteps(),
                            y=envelopes[k],
                            legendgroup=f'state({str(component)})[{k}] (envelope)',
                            name=f'state({component})[{k}] (envelope)',
                            mode='lines',
                            line={'color': color, 'dash': 'dash'},
                            showlegend=True,
                        ),
                        row=i,
                        col=1,
                    )

                    fig.append_trace(
                        Scatter(
                            x=ts,
                            y=interpolations[k],
                            legendgroup=f'state({str(component)})[{k}] (interpolated)',
                            name=f'state({component})[{k}] (interpolated)',
                            mode='lines',
                            line={'color': color, 'dash': 'dot', 'width': 0.85},
                            showlegend=True,
                            visible='legendonly',
                        ),
                        row=i,
                        col=1,
                    )

            for branch in branches:
                showlegend = True
                for i, feature in branch_features.items():
                    if feature == 'v':
                        get_interval_solutions = self.get_interval_voltages
                        get_envelope = self.get_envelope_voltages
                        get_interpolation = self.get_interpolated_voltages
                    elif feature == 'i':
                        get_interval_solutions = self.get_interval_currents
                        get_envelope = self.get_envelope_currents
                        get_interpolation = self.get_interpolated_currents
                    elif feature == 'p':
                        get_interval_solutions = self.get_interval_power
                        get_envelope = self.get_envelope_power
                        get_interpolation = self.get_interpolated_power
                    elif feature == 'c':
                        if not isinstance(branch, SwitchBranch):
                            continue

                        get_interval_solutions = lambda branch: [[1 if s else 0 for s in interval]
                                                                 for interval in self.get_interval_switch_states(branch)]
                        get_envelope = None
                        get_interpolation = None
                    else:
                        raise AssertionError

                    fig.append_trace(
                        Scatter(
                            x=list(intersperse(self._tss, None)),
                            y=list(intersperse(get_interval_solutions(branch), None)),
                            legendgroup=f'{str(component)} (simulated)',
                            name=f'{str(branch)} (simulated)',
                            mode='lines',
                            line={'color': color},
                            showlegend=showlegend,
                        ),
                        row=i,
                        col=1,
                    )

                    if get_envelope is not None:
                        fig.append_trace(
                            Scatter(
                                x=self.get_envelope_timesteps(),
                                y=get_envelope(branch),
                                legendgroup=f'{str(component)} (envelope)',
                                name=f'{str(branch)} (envelope)',
                                mode='lines',
                                line={'color': color, 'dash': 'dash'},
                                showlegend=showlegend,
                            ),
                            row=i,
                            col=1,
                        )

                    if get_interpolation is not None:
                        ts, interpolated_solutions = get_interpolation(branch)

                        fig.append_trace(
                            Scatter(
                                x=ts,
                                y=interpolated_solutions,
                                legendgroup=f'{str(component)} (interpolated)',
                                name=f'{str(branch)} (interpolated)',
                                mode='lines',
                                line={'color': color, 'dash': 'dot', 'width': 0.85},
                                showlegend=showlegend,
                                visible='legendonly',
                            ),
                            row=i,
                            col=1,
                        )

                    showlegend = False

        # Update xaxis properties
        axis_labels = {
            'v': 'voltage V [V]',
            'i': 'current I [A]',
            'p': 'power P [W]',
            's': 'state',
            'c': 'switch state',
        }

        for i, feature in enumerate(features, start=1):
            fig.update_yaxes(row=i, col=1, title_text=axis_labels[feature])

            if feature == 'c':
                fig.update_yaxes(
                    row=i,
                    col=1,
                    tickmode='array',
                    tickvals=[0, 1],
                    ticktext=['off', 'on'],
                )

        fig.update_xaxes(row=len(features), col=1, title_text="time t [s]")
        fig.update_layout(title="Transient EFM Simulation Results")

        fig.show()

    def save(self, filename: str):
        """
        Stores quickly simulation data inside the given circuit to CSV-format.

        :param filename:
            The filename to store the data in.
        """
        self.wait()

        data = [('t', self.get_timesteps())]

        # Additionally iterating over components sorts branches by component,
        # thus makes it easier for humans to read.
        for component in self._components_to_branches:
            for branch in self._components_to_branches[component]:
                data.append((f'v({str(branch)})', self.get_voltages(branch)))
                data.append((f'i({str(branch)})', self.get_currents(branch)))
                if isinstance(branch, SwitchBranch):
                    data.append((f'switchstate({str(branch)})',
                                 ['on' if ss else 'off' for ss in self.get_switch_states(branch)]))

            for i, state in enumerate(zip(*self.get_states(component)), start=1):
                data.append((f'state({str(component)})[{i}]', state))

        import csv
        with open(filename, 'w', newline='') as fl:
            w = csv.writer(fl)
            w.writerow(name for name, _ in data)
            w.writerows(zip(*(d for _, d in data)))

    def print_progress(self):
        try:
            import rich.progress
        except ImportError:
            raise RuntimeError('rich not found! Please install before using this function.')

        progress = rich.progress.Progress(
            '[progress.description]{task.description}',
            rich.progress.BarColumn(),
            '[progress.percentage]{task.percentage:>3.1f}% ({task.completed}/{task.total})',
            '•',
            'ETC',
            rich.progress.TimeRemainingColumn(),
            refresh_per_second=2,
        )

        total = len(self._skips) + 1
        with progress:
            task = progress.add_task(
                'simulate (EFM)',
                total=total,
                completed=len(self._tss),
            )

            while not self._finished.wait(0.5):
                progress.update(task, completed=len(self._tss))

            progress.update(task, completed=total)



from dataclasses import dataclass
from typing import Optional, Callable

import numpy as np

from .Switch import Switch


@dataclass(eq=False)
class ComplementarySwitch(Switch):
    """
    A convenience class that acts complementary to another specified switch.

    switch:
        The switch to generate the complementary signal for.
    """
    switch: Switch

    def switch_state(self, t: float) -> bool:
        return not self.switch.switch_state(t)

    def next_event(self, vi: Callable[[float], np.ndarray], t1: float, t2: float) -> Optional[float]:
        return self.switch.next_event(vi, t1, t2)



from __future__ import annotations

from dataclasses import dataclass, field
from math import exp
from typing import Hashable, List, Tuple, Sequence

import numpy as np

from .Branch import Branch, CurrentBranch
from .Component import Component


@dataclass(eq=False)
class BipolarJunctionTransistor(Component):
    """
    Describes a bipolar junction transistor.

    This component uses the simplified large-signal Ebers-Moll model
    (see https://en.wikipedia.org/wiki/Bipolar_junction_transistor#Ebers%E2%80%93Moll_model).

    This component is connected in B-C-E order. The first given terminal connects to the base, the second one to
    the collector and the last one to the emitter.

    For convenience you can immediately access the currents at the B/C/E terminals by using the `i_b`, `i_c` and `i_e`
    properties respectively. Mind their current direction. Also you can access `v_cb` (C -> B), `v_be` (B -> E)  and
    `v_ce` (C -> E) to get the voltage drops.

    i_es:
        Reverse saturation current I_ES of the base–emitter diode (on the order of 10^−15 to 10^−12 amperes).
    alpha_f:
        Common base forward short-circuit current gain (usually 0.98 to 0.998)
    v_t:
        Thermal voltage V_T. For approximate room temperate 300K this value is 25.85mV (0.02585V). This is default.
    """

    @dataclass(eq=False)
    class CollectorBaseBranch(CurrentBranch):
        def get_current(self, v_i: np.ndarray, t1: float, t2: float) -> float:
            return self.component.i_es * (exp(v_i[1] / self.component.v_t) - 1) * self.component.alpha_f

        def get_jacobian(self, v_i: np.ndarray, t1: float, t2: float) -> Sequence[float]:
            return [
                0,
                self.component.i_es / self.component.v_t * exp(v_i[1] / self.component.v_t) * self.component.alpha_f
            ]

    @dataclass(eq=False)
    class BaseEmitterBranch(CurrentBranch):
        def get_current(self, v_i: np.ndarray, t1: float, t2: float) -> float:
            return self.component.i_es * (exp(v_i[1] / self.component.v_t) - 1)

        def get_jacobian(self, v_i: np.ndarray, t1: float, t2: float) -> Sequence[float]:
            return [0, self.component.i_es / self.component.v_t * exp(v_i[1] / self.component.v_t)]

    i_es: float
    alpha_f: float
    v_t: float = 0.02585
    _collector_base_branch: CollectorBaseBranch = field(init=False, repr=False)
    _base_emitter_branch: BaseEmitterBranch = field(init=False, repr=False)

    def __post_init__(self):
        self._collector_base_branch = self.CollectorBaseBranch(self)
        self._collector_base_branch.name = 'CB'
        self._base_emitter_branch = self.BaseEmitterBranch(self)
        self._base_emitter_branch.name = 'BE'

    def connect(self,
                base: Hashable,
                collector: Hashable,
                emitter: Hashable) -> List[Tuple[Hashable, Hashable, Branch]]:
        """
        Connects the transistor.

        :param base:
            The base (B) terminal.
        :param collector:
            The collector (C) terminal.
        :param emitter:
            The emitter (E) terminal.
        """
        return [
            (collector, base, self._collector_base_branch),
            (base, emitter, self._base_emitter_branch),
        ]



from dataclasses import dataclass
from math import pi
from typing import Optional, Callable

import numpy as np

from .Switch import Switch


@dataclass(eq=False)
class PeriodicSwitch(Switch):
    """
    A switch with a periodic on/off-pattern.

    This class is a building block to define your own simple switches easily.
    Subclasses override the `get_switch_signal()` function that supplies the switch states inside a single period.

    frequency:
        The current frequency (measured in Hertz).
    phase:
        The initial phase angle (measured in rad).
    """
    frequency: float = 1.0
    phase: float = 0.0

    @property
    def T(self):
        """
        The period time :math:`T`.
        """
        return 1.0 / self.frequency

    @T.setter
    def T(self, value):
        self.frequency = 1.0 / value

    def get_phase_from_time(self, t):
        return t * 2 * pi * self.frequency + self.phase

    def switch_state(self, t: float) -> bool:
        return self.get_switch_signal(self.get_phase_from_time(t) % (2 * pi))

    def get_switch_signal(self, phase: float) -> bool:
        """
        Returns the signal.

        :param phase:
            The momentary angular phase, inside the range [0, 2pi).
        :return:
            The switch state.
        """
        raise NotImplementedError

    def next_event(self, vi: Callable[[float], np.ndarray], t1: float, t2: float) -> Optional[float]:
        interval1, phase1 = divmod(self.get_phase_from_time(t1), 2 * pi)
        interval2, phase2 = divmod(self.get_phase_from_time(t2), 2 * pi)

        intv_phase1 = phase1
        for p in range(int(interval1), int(interval2)):
            intv_phase2 = np.nextafter(2 * pi, 0)

            eventphase = self.next_switch_event(intv_phase1, intv_phase2)
            if eventphase is not None and 0 <= eventphase < 2 * pi:
                return (p + (eventphase - self.phase) / (2 * pi)) / self.frequency

            intv_phase1 = 0

        eventphase = self.next_switch_event(intv_phase1, phase2)
        return None if eventphase is None else (int(interval2) + (eventphase - self.phase) / (2 * pi)) / self.frequency

    def next_switch_event(self, phase1: float, phase2: float) -> Optional[float]:
        r"""
        Calculates the next switch event within a phase interval.

        Instead of event times, phases indicating the switch event are returned.
        They are properly translated to according simulation times.

        This function guarantees to serve phases in the range :math:`[0, 2\pi)`.

        :param phase1:
            Previous phase (measured in rad).
        :param phase2:
            Present phase (measured in rad).
        :return:
            The next phase at which a switch state change occurs, or None if nothing happens.
        """
        return None



from dataclasses import dataclass
from math import pi, sqrt
from typing import Optional

from .PeriodicCurrentSource import PeriodicCurrentSource


@dataclass(eq=False)
class CurrentSourceRectangular(PeriodicCurrentSource):
    """
    A rectangular current source supply.

    By default the current toggles between `+amplitude` and `-amplitude`.
    To make the signal toggle between `0` and `a`, set `offset = amplitude = a / 2`.

    By default the duty cycle of the source is 0.5, meaning it toggles equally between min and max.
    The duty cycle determines how much time (relative) is spent between the min and max states.
    A value of 0 or below means the source becomes effectively a DC source serving constantly the minimum
    voltage. A value of 1 or above is effectively a constant DC source at maximum voltage.
    For example a value in between of 0.75 means that 75% of the period is spent at maximum voltage
    and the following 25% is minimum voltage.

    amplitude:
        The amplitude of the rectangular current (measured in Amperes).
    offset:
        The current offset of the signal (measured in Amperes).
    duty:
        The duty cycle. Meaningful values lie between 0 and 1.
    """
    amplitude: float = 1.0
    offset: float = 0.0
    duty: float = 0.5

    @property
    def effective_amplitude(self):
        r"""
        The effective amplitude (:math:`A_{RMS}`) depends not only on the amplitude, but also on the offset and duty
        cycle.
        According to the formula for RMS (see https://en.wikipedia.org/wiki/Root_mean_square and
        https://en.wikipedia.org/wiki/RMS_amplitude)

        .. math::

             A_{RMS} &= \sqrt{\frac{1}{T} \int_{0}^{T}{f^2(t) dt}} \\
                     &= \sqrt{
                            \frac{1}{T}
                                \left (
                                    \int_{0}^{\delta T}{(o + A)^2 dt} +
                                    \int_{\delta T}^{T}{(o - A)^2 dt}
                                \right )
                            } \\
                     &= \sqrt{
                            (o + A)^2 \delta T +
                            (o - A)^2 (1 - \delta) T
                        }

        :return:
            The effective amplitude current (measured in Amperes).
        """
        return sqrt(
            (self.offset + self.amplitude)**2 * self.duty +
            (self.offset - self.amplitude)**2 * (1 - self.duty)
        )

    def get_signal(self, phase: float) -> float:
        high = phase < 2 * pi * self.duty
        return self.amplitude * (1 if high else -1) + self.offset

    def next_signal_event(self, phase1: float, phase2: float) -> Optional[float]:
        if phase1 == 0.0:
            return 0.0

        dutyphase = self.duty * 2 * pi
        if phase1 <= dutyphase <= phase2:
            return dutyphase

        return None



from .BipolarJunctionTransistor import BipolarJunctionTransistor
from .Branch import Branch, CurrentBranch, VoltageBranch, SwitchBranch
from .Capacitance import Capacitance, C
from .ComplementarySwitch import ComplementarySwitch
from .Component import Component
from .CoupledInductance import CoupledInductance
from .CurrentSourceAC import CurrentSourceAC
from .CurrentSourceDC import CurrentSourceDC
from .CurrentSourceHalfWaveAC import CurrentSourceHalfWaveAC
from .CurrentSourcePWM import CurrentSourcePWM
from .CurrentSourceRectangular import CurrentSourceRectangular
from .Inductance import Inductance, L
from .LinearizedShockleyDiode import LinearizedShockleyDiode
from .Motor import Motor
from .OperationalAmplifier import OperationalAmplifier
from .PeriodicSwitch import PeriodicSwitch
from .Resistance import Resistance, R
from .ShockleyDiode import ShockleyDiode
from .Switch import Switch
from .SwitchPWM import SwitchPWM
from .SwitchRectangular import SwitchRectangular
from .TwoTerminalCurrentComponent import TwoTerminalCurrentComponent
from .TwoTerminalVoltageComponent import TwoTerminalVoltageComponent
from .VoltageSourceAC import VoltageSourceAC
from .VoltageSourceDC import VoltageSourceDC
from .VoltageSourceHalfWaveAC import VoltageSourceHalfWaveAC
from .VoltageSourcePWM import VoltageSourcePWM
from .VoltageSourceRectangular import VoltageSourceRectangular



from typing import Callable

import numpy as np
from scipy.optimize import root, OptimizeResult

from respice.math.optimization.rootsolvers.newton import solve as newton


def solve(eq: Callable,
          x0: np.ndarray,
          args=tuple(),
          jac=None,
          hybr_args=None,
          newton_args=None,
          lm_args=None) -> OptimizeResult:

    if hybr_args is None:
        hybr_args = {}

    try:
        result = root(eq, x0, args=args, method='hybr', jac=jac, **hybr_args)
        if result.success:
            result.step = 'hybr'
            return result
    except OverflowError:
        pass  # The newton solver is more robust against overflows.

    if jac is not None:
        if newton_args is None:
            newton_args = {}

        f = eq if jac is True else lambda x, *args: (eq(x, *args), jac(x, *args))
        result = newton(f, x0, args=args, **newton_args)
        if result.success:
            result.step = 'newton'
            return result

    if lm_args is None:
        lm_args = {}

    result = root(eq, x0, args=args, method='lm', jac=jac, **lm_args)
    result.step = 'lm'
    return result



from functools import reduce
from math import factorial

import numpy.polynomial.polynomial as poly


def _extend_add(l, r):
    r[:len(l)] += l
    return r


class TaylorPolynomial:
    def __init__(self, fs, a):
        r"""
        Creates a new taylor polynomial from given function and derivative values.

        :param fs:
            Function and derivative values: :math:`f(x), f'(x), f''(x), \cdots`
        :param a:
            Expansion point.
        """
        self._c = reduce(_extend_add, (f / factorial(i) * poly.polyfromroots([a] * i)
                                       for i, f in enumerate(fs)))

    def __call__(self, x):
        return poly.polyval(x, self._c)

    def derivate(self, m=1):
        return poly.Polynomial(poly.polyder(self._c, m))



from dataclasses import dataclass

from .TwoTerminalCurrentComponent import TwoTerminalCurrentComponent


@dataclass(eq=False)
class CurrentSourceDC(TwoTerminalCurrentComponent):
    """
    A DC current supply.

    Current emits in the same direction as the edge points in the circuit graph.

    value:
        The current emitted by this supply (measured in Amperes).
    """
    value: float

    def get_current(self, v: float, t1: float, t2: float) -> float:
        return self.value

    def get_jacobian(self, v: float, t1: float, t2: float) -> float:
        return 0.0



from respice.analysis import Circuit
from respice.components import Component


def create_series_connected_circuit(*components: Component) -> Circuit:
    """
    Utility to connect a closed circuit with all given elements connected in series.

    :param components:
        The elements to construct the circuit from.
    :return:
        A circuit.
    """
    if len(components) < 2:
        raise ValueError('Not enough elements to connect in series. At least 2 required.')

    circuit = Circuit()

    for i, elem in enumerate(components[:-1]):
        circuit.add(elem, i, i + 1)

    # Add closing element.
    circuit.add(components[-1], len(components) - 1, 0)

    return circuit



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Hashable, List, Tuple, Sequence

import numpy as np

from .Branch import Branch, VoltageBranch, CurrentBranch
from .Component import Component


@dataclass(eq=False)
class CoupledInductance(Component):
    """
    Represents an ideally mutually coupled coil.

    n1:
        The number of windings on coil 1 ("left-hand").
    n1:
        The number of windings on coil 2 ("right-hand").
    """

    @dataclass(eq=False)
    class CoupledInductanceCoil1(VoltageBranch):
        def get_voltage(self, v_i: np.ndarray, t1: float, t2: float) -> float:
            return v_i[1] * self.component.n1 / self.component.n2

        def get_jacobian(self, v_i: np.ndarray, t1: float, t2: float) -> Sequence[float]:
            return [0, self.component.n1 / self.component.n2]

    @dataclass(eq=False)
    class CoupledInductanceCoil2(CurrentBranch):
        def get_current(self, v_i: np.ndarray, t1: float, t2: float) -> float:
            return v_i[0] * self.component.n1 / self.component.n2

        def get_jacobian(self, v_i: np.ndarray, t1: float, t2: float) -> Sequence[float]:
            return [self.component.n1 / self.component.n2, 0]

    n1: int
    n2: int
    coil1: CoupledInductanceCoil1 = field(init=False, repr=False)
    coil2: CoupledInductanceCoil2 = field(init=False, repr=False)

    def __post_init__(self):
        self.coil1 = self.CoupledInductanceCoil1(self)
        self.coil1.name = 'coil1'
        self.coil2 = self.CoupledInductanceCoil2(self)
        self.coil1.name = 'coil2'

    def connect(self,
                coil1_terminal1: Hashable,
                coil1_terminal2: Hashable,
                coil2_terminal1: Hashable,
                coil2_terminal2: Hashable) -> List[Tuple[Hashable, Hashable, Branch]]:
        """
        Connects the mutually coupled inductance.

        :param coil1_terminal1:
            Terminal 1 of the first coil.
        :param coil1_terminal2:
            Terminal 2 of the first coil.
        :param coil2_terminal1:
            Terminal 1 of the second coil.
        :param coil2_terminal2:
            Terminal 2 of the second coil.
        """
        return [
            (coil1_terminal1, coil1_terminal2, self.coil1),
            (coil2_terminal1, coil2_terminal2, self.coil2),
        ]



import sys
from bisect import insort
from dataclasses import dataclass
from itertools import compress, chain
from threading import Thread
from types import SimpleNamespace
from typing import Hashable, List, Set, Collection, Union

import numpy as np
from networkx import Graph, connected_components, incidence_matrix
from scipy.sparse.linalg import spsolve

from respice.cachetools import LRUCache
from respice.collections import OneToManyInvertibleMapping
from respice.components import Component, CurrentBranch, VoltageBranch, SwitchBranch
from respice.itertools.custom import compact, uncompact
from respice.math import flcm
from respice.math.optimization.rootsolvers.mixed import solve
from respice.threading import prefer_thread_yield
from .MNAEquationStack import MNAEquationStack
from .TransientEFMSimulation import TransientEFMSimulation
from .TransientSimulation import TransientSimulation
from .TransientSteadyStateEFMSimulation import TransientSteadyStateEFMSimulation
from .TransientSteadyStateSimulation import TransientSteadyStateSimulation
from .UniqueEdgeMultiDiGraph import UniqueEdgeMultiDiGraph


def linear_interpolation(x1, ys1, x2, ys2):
    if x2 - x1 == 0.0:
        return constant_interpolation(ys2)

    # Do not use numpy.polyfit to perform this task! It is about an order of magnitude slower.
    # It is actually tailored for least square fitting, so it prepares much more data under the hood which takes time.
    polys = [np.poly1d([(y2 - y1) / (x2 - x1), (y1 * x2 - x2 * x1) / (x2 - x1)])
             for y1, y2 in zip(ys1, ys2)]
    return lambda v: np.fromiter((poly(v) for poly in polys), dtype=float)


def constant_interpolation(y):
    return lambda v: y


@dataclass(eq=False)
class _MergedNode:
    nodes: Set[Hashable]


class ConvergenceError(Exception):
    pass


class Circuit:
    _EVENT_RESAMPLING_PRECISION = sys.float_info.epsilon * 2**20
    _CIRCUIT_SWITCH_CONFIGURATION_CACHE_SIZE = 2 ** 16

    def __init__(self):
        self._graph = UniqueEdgeMultiDiGraph()

        # Enforces uniqueness of elements across all graph edges (additionally, since uniqueness is effectively already
        # enforced by the special MultiDiGraph type used, because same branches of components can't be added twice as
        # well). Since adding the same component twice would cause the data in its branches to become mixed up, since
        # multiple results get appended to the same branch instance.
        self._component_branches = OneToManyInvertibleMapping()

        self._cache = SimpleNamespace()
        self._cache.circuit_switch_configurations = LRUCache(maxsize=self._CIRCUIT_SWITCH_CONFIGURATION_CACHE_SIZE)

    def add(self, component: Component, *terminals: Hashable):
        """
        Adds a new component to the circuit.

        :param component:
            The component to connect.
        :param terminals:
            The nodes/potentials to connect the component to. See each component's documentation about the order of
            each terminal to be specified.
        """
        if component in self._component_branches:
            raise KeyError(f'given component {repr(component)} already added to circuit.')

        component_branches = component.connect(*terminals)

        for source, target, branch in component_branches:
            self._component_branches[component] = branch
            self._graph.add_edge(source, target, branch)

        self._cache.circuit_switch_configurations.clear()

    def remove(self, component: Component):
        """
        Removes a component from the circuit.

        :param component:
            The component to remove from the circuit.
        """
        for branch in self._component_branches[component]:
            self._graph.remove_edge(branch)

        del self._component_branches[component]

        self._cache.circuit_switch_configurations.clear()

    @property
    def components(self):
        """
        :return:
            All components added to this circuit.
        """
        return self._component_branches.keys()

    def _get_coupled_vi_vector(self, component, result, eq):
        coupled_v_i = []
        for coupled_branch in self._component_branches[component]:
            if isinstance(coupled_branch, CurrentBranch):
                coupled_v_i.append(eq.get_voltage(result, coupled_branch))
            elif isinstance(coupled_branch, VoltageBranch):
                coupled_v_i.append(eq.get_current(result, coupled_branch))
            elif isinstance(coupled_branch, SwitchBranch):
                pass
            else:
                raise AssertionError(f'Unknown branch type encountered: f{type(coupled_branch)}')

        return np.array(coupled_v_i)

    def _get_switches(self):
        # Ensures the same order of switches between calls so switch configurations can be stored reliably.
        return [branch for branch in self._component_branches.reverse if isinstance(branch, SwitchBranch)]

    def simulate(self, ts1: float, ts2: float, steps: int, preferred_ground_nodes=frozenset()) -> TransientSimulation:
        """
        Perform a circuit simulation.

        The simulation is *asynchronous*! A future object is returned that allows you to access results and progress
        during simulation runtime. Note the following:

        * Accesses to single component or branch values (such as voltage, current, etc.) will immediately return the
          most recent simulation data. Between calls, the data may change. For safe access, although not really intended
          to be utilized by users, you can acquire the modification lock by:

          .. code-block:: python

              simulation = circuit.simulate(...)
              with simulation as sim:
                  # Acquired the simulation lock. The simulation will halt when attempting to write results.
                  # This means be sure to release it again if you don't want to interrupt the simulation too long.
                  pass

        * Accessing properties and fields of components or branches is safe, but writing is not and will consequently
          influence the simulation and causes undefined behavior.

        * Cancelling a simulation via `KeyboardInterrupt` alias `Ctrl + C` is not handled automatically. If you want to
          support aborting simulations, wrap the simulation code in a try-except and cancel it:

          .. code-block:: python

              simulation = circuit.simulate(...)
              try:
                  simulation.wait()
              except KeyboardInterrupt:
                  simulation.cancel()

        States after a simulation remain. This means you can chain simulations
        together altering parameters in between (for example turning off a voltage supply by removing
        it from the circuit or changing a resistance value).

        :param ts1:
            Initial point in time to start simulation from.
        :param ts2:
            Point in time to simulate up to.
        :param steps:
            The number of steps to simulate between `t1` and `t2` (`t1` and `t2` included).
        :param preferred_ground_nodes:
            Notes to prefer as ground (GND).
        :return:
            A `TransientSimulation` to access the result of the simulation from.
            Note that computation happens *asynchronously*!
        """
        if steps < 1:
            raise ValueError('at least 1 step required for simulation')

        simulation = TransientSimulation(ts1, ts2, self._component_branches, self._graph.nodes)

        def simulate_async(simulation, ts1, ts2, steps, preferred_ground_nodes):
            try:
                self._simulate(simulation, ts1, ts2, steps, preferred_ground_nodes)
            except Exception as ex:
                simulation.abort(ex)

        Thread(target=simulate_async,
               args=(simulation, ts1, ts2, steps, preferred_ground_nodes)).start()

        return simulation

    def _simulate(self,
                  simulation: TransientSimulation,
                  ts1: float,
                  ts2: float,
                  steps: int,
                  preferred_gnd_nodes=frozenset()):
        # A standard list is explicitly needed to allow modifications during iteration.
        ts = list(np.linspace(ts1, ts2, steps))

        # Closures for less code duplication.

        def create_switch_configured_graph(switch_config):
            # Generate a graph representation that can be fed into MNAEquationStack.
            # MNAEquationStack is not able to handle switches by itself, so graphs are "configured", which means that
            # off-switches are plainly removed from the graph and adjacent nodes of on-switches are merged together
            # effectively symbolizing a short-circuit between those nodes.
            switch_configured_graph: UniqueEdgeMultiDiGraph = self._graph.copy()
            merged_nodes = Graph()  # Use a graph to relate nodes for later merge.
            for switch, switch_on in zip(self._get_switches(), switch_config):
                if switch_on:
                    merged_nodes.add_edge(*self._graph.get_nodes(switch))

                switch_configured_graph.remove_edge(switch)

            for nodes in connected_components(merged_nodes):
                switch_configured_graph.merge_nodes(nodes, _MergedNode(nodes))

            return switch_configured_graph

        def solve_circuit(eq1, eq2, x0, t1, t2):
            if eq1 is not eq2:
                # Convert old result/guess to new guess.
                x0 = eq2.assemble_vector(*eq1.disassemble_vector(x0))

            # eq2 is the equation to be solved for current time t2.
            result = solve(eq2.lambdify(), x0, args=(t1, t2), jac=eq2.lambdify_jacobian())

            if result.success:
                return result.x
            else:
                raise ConvergenceError(result)

        def get_equation(t):
            # Get the correct circuit configuration and according equations at time t.

            switch_config = tuple(switch.switch_state(t) for switch in self._get_switches())

            if switch_config in self._cache.circuit_switch_configurations:
                switch_configured_graph, equation = self._cache.circuit_switch_configurations[switch_config]
            else:
                try:
                    # Prefer ground nodes from a previous simulation.
                    _, (_, eq) = self._cache.circuit_switch_configurations.last()
                    gnd_nodes = set(eq.get_reference_nodes())
                except KeyError:
                    gnd_nodes = set()

                gnd_nodes |= preferred_gnd_nodes
                switch_configured_graph = create_switch_configured_graph(switch_config)

                equation = MNAEquationStack(
                    switch_configured_graph,
                    ([branch
                      for branch in coupled_branches
                      if not isinstance(branch, SwitchBranch)]
                     for coupled_branches in self._component_branches.values()),
                    gnd_nodes)
                self._cache.circuit_switch_configurations[switch_config] = switch_configured_graph, equation

            return switch_config, switch_configured_graph, equation

        def store_result(t1, t2, eq, result, switch_config, switch_configured_graph):
            with simulation as sim:
                sim.ts.append(t2)

                # Store node->voltages map.
                for node, v in eq.get_node_voltages(result).items():
                    if isinstance(node, _MergedNode):
                        for merged_node in node.nodes:
                            sim.node_potentials[merged_node].append(v)
                    else:
                        sim.node_potentials[node].append(v)

                # Store switch states.
                for switch, switch_state in zip(self._get_switches(), switch_config):
                    sim.switch_states[switch].append(switch_state)

                # Map back calculated voltages and currents to actual elements.
                for component in self._component_branches:
                    coupled_v_i = self._get_coupled_vi_vector(component, result, eq)

                    for branch in self._component_branches[component]:
                        if isinstance(branch, CurrentBranch):
                            v = eq.get_voltage(result, branch)
                            i = branch.get_current(coupled_v_i, t1, t2)
                            sim.branch_voltages[branch].append(v)
                            sim.branch_currents[branch].append(i)
                        elif isinstance(branch, VoltageBranch):
                            i = eq.get_current(result, branch)
                            v = branch.get_voltage(coupled_v_i, t1, t2)
                            sim.branch_voltages[branch].append(v)
                            sim.branch_currents[branch].append(i)
                        elif isinstance(branch, SwitchBranch):
                            # If the switch is on, it produces a merged node. This scenario is handled below.
                            if not branch.switch_state(t2):
                                source, target = self._graph.get_nodes(branch)
                                sim.branch_voltages[branch].append(sim.node_potentials[target][-1] -
                                                                   sim.node_potentials[source][-1])
                                sim.branch_currents[branch].append(0.0)
                        else:
                            raise AssertionError(f'Unknown branch type encountered: f{type(branch)}')

                    component.update(coupled_v_i, t1, t2)

                    # Save state results now that components are updated.
                    sim.component_states[component].append(component.state)

                # Calculate switch branch currents.
                for node in switch_configured_graph.nodes:
                    if isinstance(node, _MergedNode):
                        # Use the incidence matrix to solve for any kind of switch layout.
                        # If especially switches are connected to other switches, with the help of the
                        # incidence matrix it is possible to solve for all currents through those switches.
                        merged_graph = self._graph.subgraph(node.nodes).copy()

                        # Sometimes other branches than switches can form self loops if they are connected between
                        # the switches of the same merged node. In this case we need to exclude those from the
                        # incidence matrix.
                        merged_graph.remove_edges_from({branch
                                                        for _, _, branch in merged_graph.edges
                                                        if not isinstance(branch, SwitchBranch)})

                        nodes = list(merged_graph.nodes)[:-1]

                        A = incidence_matrix(merged_graph, oriented=True)[:-1]
                        bi = np.fromiter(
                            (sum(sim.branch_currents[branch][-1]
                                 for _, _, branch in (set(self._graph.in_edges(node, keys=True)) -
                                                      set(merged_graph.in_edges(node, keys=True))))
                             for node in nodes),
                            dtype=float,
                        )
                        bo = np.fromiter(
                            (sum(sim.branch_currents[branch][-1]
                                 for _, _, branch in (set(self._graph.out_edges(node, keys=True)) -
                                                      set(merged_graph.out_edges(node, keys=True))))
                             for node in nodes),
                            dtype=float,
                        )

                        switch_currents_solutions = spsolve(A, bo - bi)
                        for i, (_, _, branch) in zip(switch_currents_solutions, merged_graph.edges(keys=True)):
                            sim.branch_voltages[branch].append(0.0)
                            sim.branch_currents[branch].append(i)

        def get_events(t1, t2, xinterp):
            return [event
                    for event in (
                        component.next_event(xinterp(component), t1, t2)
                        for component in self._component_branches)
                    if event is not None and t1 <= event <= t2]

        def get_next_event(t1, t2, xinterp):
            events = get_events(t1, t2, xinterp)
            return min(events) if events else None

        titer = iter(ts)
        t1 = next(titer)

        # Specially solve for t0.
        sc1, sc1graph, eq1 = get_equation(t1)
        x0 = np.zeros(len(eq1))
        result1 = solve_circuit(eq1, eq1, x0, t1, t1)
        store_result(t1, t1, eq1, result1, sc1, sc1graph)
        # Event handling is special in the region [t0, t0+ERP].
        # Events before t0 must be considered too, since they resolve inside the desired calculation range.
        # All events' (that have an influence, thus (t0-ERP, t0+ERP)) post-resolution points must be staged for
        # analysis.
        # Although less accurrate, for performance the state is assumed to be the result at t0.
        # It is possible to back-compute, but not worth the effort considering the interval is very small
        # (although as of the time writing this, all components use the trapezoidal integration rule which is symmetric
        # in this regard).
        for event in get_events(np.nextafter(t1 - self._EVENT_RESAMPLING_PRECISION, np.inf),
                                np.nextafter(t1 + self._EVENT_RESAMPLING_PRECISION, -np.inf),  # This is not a typo, it is t1.
                                lambda component: constant_interpolation(
                                    self._get_coupled_vi_vector(component, result1, eq1)
                                )):
            insort(ts, event + self._EVENT_RESAMPLING_PRECISION)

        t2 = next(titer)
        # Due to event handling, additional points after the desired range may be inserted, which make the loop
        # condition necessary.
        while t2 <= ts2:
            if simulation.is_cancelled:
                return simulation

            if t1 == t2:
                # Sometimes event times and externally desired time steps fall exactly on the same spot. In this case,
                # skip the redundant simulation.
                try:
                    t2 = next(titer)
                except StopIteration:
                    break

            sc2, sc2graph, eq2 = get_equation(t2)
            result2 = solve_circuit(eq1, eq2, result1, t1, t2)

            te1 = np.nextafter(t1 + self._EVENT_RESAMPLING_PRECISION, np.inf)
            te2 = np.nextafter(t2 + self._EVENT_RESAMPLING_PRECISION, -np.inf)

            event = get_next_event(
                te1,
                te2,
                lambda component: linear_interpolation(
                    t1,
                    self._get_coupled_vi_vector(component, result1, eq1),
                    t2,
                    self._get_coupled_vi_vector(component, result2, eq2),
                ),
            )

            if event is None:
                store_result(t1, t2, eq2, result2, sc2, sc2graph)
                t1 = t2
                try:
                    t2 = next(titer)
                except StopIteration:
                    break
            else:
                te1 = event - self._EVENT_RESAMPLING_PRECISION
                te2 = event + self._EVENT_RESAMPLING_PRECISION

                sc2, sc2graph, eq2 = get_equation(te1)
                result2 = solve_circuit(eq1, eq2, result1, t1, te1)

                store_result(t1, te1, eq2, result2, sc2, sc2graph)

                if te2 < t2:
                    t2 = te2

                insort(ts, te2)  # This operation has no considerable side-effects on the iterator.

                t1 = te1

            result1 = result2
            eq1 = eq2
            sc1graph = sc2graph

            # To allow cooperative programming, prefer to yield the thread here after a certain runtime.
            prefer_thread_yield()

        simulation.finish()
        return simulation

    def simulate_efm(self,
                     t0: float,
                     T: float,
                     interval_steps: int,
                     skips: List[int],
                     preferred_ground_nodes: Set = frozenset()) -> TransientEFMSimulation:
        """
        Simulates the circuit using the envelope-following-method (EFM).

        The simulation is *asynchronous*! A future object is returned that allows you to access results and progress
        during simulation runtime. The same notes as for the standard `simulate` apply.

        The EFM is beneficial to analyze systems faster with the trade-off of accuracy.
        With appropriate parameters and a suitable circuit this method can simulate significantly faster.
        This is especially beneficial for multi-rate systems with highly different intrinsical frequencies.

        The EFM requires you to know the intrinsical period time T where jumps can be properly made from.
        After that you define how many steps a single interval is simulated with and how many skips (`m`) should be
        performed.
        With this information the EFM performs skips according to following implicit equation:

        .. math::

            f(t + m T) = f(t) + \\frac12 m [f(t+T) - f(t) + f(t + (m+1) T) - f(t + mT)]

        The solution is then the so called "envelope", the guiding solution of the system.

        You can choose arbitrary consecutive skips organized in a list. If you simply want constant skipping, you can
        do `[s] * n` instead of writing out `[s, s, ..., s]`.

        For reference, see the paper at https://ieeexplore.ieee.org/document/4342106.

        :param t0:
            The initial time to start simulation from.
        :param T:
            Period time of a single interval.
        :param interval_steps:
            How many steps to simulate for a single interval.
        :param skips:
            Consecutive skips to perform encoded in a list.
        :param preferred_ground_nodes:
            Notes to prefer as ground (GND).
        :return:
            A `TransientEFMSimulation` object containing the result of the simulation.
            Note that computation happens *asynchronously*!.
        """

        if len(skips) <= 0:
            raise ValueError('skips must contain at least one value')
        if interval_steps < 1:
            raise ValueError('interval_steps must be greater than 1')

        simulation = TransientEFMSimulation(skips, self._component_branches, self._graph.nodes)

        def simulate_efm_async(simulation, t0, T, interval_steps, skips, preferred_ground_nodes):
            try:
                self._simulate_efm(simulation, t0, T, interval_steps, skips, preferred_ground_nodes)
            except Exception as ex:
                simulation.abort(ex)

        Thread(target=simulate_efm_async,
               args=(simulation, t0, T, interval_steps, skips, preferred_ground_nodes)).start()

        return simulation

    def _simulate_efm(self,
                      simulation: TransientEFMSimulation,
                      t0: float,
                      T: float,
                      interval_steps: int,
                      skips: List[int],
                      preferred_ground_nodes: Set = frozenset()):

        # shorthand prefixes for variables: li -> left interval, ri -> right interval

        li_ts1 = t0
        # s corresponds to m - 2 in the backing paper.
        for s in skips:
            if simulation.is_cancelled:
                return simulation

            li_ts2 = li_ts1 + T

            li_state1, state_compactification = compact([component.state for component in self.components])
            li_state1 = np.array(li_state1)
            subsim = self.simulate(li_ts1, li_ts2, interval_steps, preferred_ground_nodes)
            subsim.wait()
            li_state2, _ = compact([component.state for component in self.components])
            li_state2 = np.array(li_state2)
            li_state_delta = li_state2 - li_state1

            with simulation as sim:
                sim.append_from_simulation(subsim)

            ri_ts1 = li_ts2 + s * T
            ri_ts2 = ri_ts1 + T

            def efm_equation(ri_state1):
                for component, state in zip(self.components, uncompact(ri_state1, state_compactification)):
                    component.state = state

                ri_state1 = np.array(ri_state1)
                self.simulate(ri_ts1, ri_ts2, interval_steps, preferred_ground_nodes).wait()
                ri_state2, _ = compact([component.state for component in self.components])
                ri_state2 = np.array(ri_state2)
                ri_state_delta = ri_state2 - ri_state1

                # The equations are an adjusted version with better solution consistency.
                # Instead of calculating up to the end of the second interval, the efm-equations have to hold only up to
                # the beginning of the second interval. Effectively, m was changed from the original m to m - 1.
                return li_state2 - ri_state1 + s * 0.5 * (li_state_delta + ri_state_delta)

            initial_guess = li_state2 + s * li_state_delta  # Initial guess by forward euler approximation.
            satisfying_state = solve(efm_equation, initial_guess).x

            states = uncompact(satisfying_state, state_compactification)
            for component, state in zip(self.components, states):
                component.state = state

            li_ts1 = ri_ts1

            # To allow cooperative programming, prefer to yield the thread here after a certain runtime.
            prefer_thread_yield()

        # With steps > 0 being enforced above, referencing i won't cause any problems. The loop runs at least once.
        li_ts2 = li_ts1 + T
        subsim = self.simulate(li_ts1, li_ts2, interval_steps, preferred_ground_nodes)
        subsim.wait()

        with simulation as sim:
            sim.append_from_simulation(subsim)

        simulation.finish()
        return simulation

    def steadystate(self, T, steps, t0=0, preferred_ground_nodes: Set = frozenset()) -> TransientSteadyStateSimulation:
        """
        Finds the periodic steady-state solution of the circuit.

        The simulation is *asynchronous*! A future object is returned that allows you to access results and progress
        during simulation runtime. The same notes as for the standard `simulate` apply.

        Periodic steady-states occur for example for all linear circuits where complex analysis can be applied
        (i.e. circuits with sinusoidal sources, resistors, capacitors, inductors).
        This function requires you to pass ahead a suitable period `T` that denotes the circuit system's period time.
        For multi-rate circuits (i.e. with many sinusoidal sources), `T` becomes the least common multiple of all
        those period times Ts (the least common multiple is here extended to rational numbers - for non-rational
        frequencies, T can be approximated by rationalizing).

        The steady-state result is stored as a single period simulation inside the circuit's components and can be
        accessed exactly like with `simulate`. If the state-solution itself is of interest, respective states can
        be queried by using `component.state`.

        :param T:
            The circuit system's characteristic period time `T`.
        :param steps:
            Steps to simulate per period `T`.
            For sinusoidal sources, can be approximately set to `10 * T / min(Ts)`.
        :param t0:
            Initial time to start simulation from.
        :param preferred_ground_nodes:
            Notes to prefer as ground (GND).
        :return:
            A simulation object.
            Note that computation happens *asynchronously*!
        """
        simulation = TransientSteadyStateSimulation(t0, t0 + T, self._component_branches, self._graph.nodes)

        def steadystate_async(simulation, T, steps, t0, preferred_ground_nodes):
            try:
                self._steadystate(simulation, T, steps, t0, preferred_ground_nodes)
            except Exception as ex:
                simulation.abort(ex)

        Thread(target=steadystate_async,
               args=(simulation, T, steps, t0, preferred_ground_nodes)).start()

        return simulation

    def _steadystate(self,
                     simulation: TransientSteadyStateSimulation,
                     T: float,
                     steps: int,
                     t0=0,
                     preferred_ground_nodes=frozenset()):
        state0, state_compactification = compact([component.state for component in self.components])

        ts1 = t0
        ts2 = t0 + T

        def problem(compacted_state):
            states = uncompact(compacted_state, state_compactification)

            for component, state in zip(self.components, states):
                component.state = state

            self.simulate(ts1, ts2, steps, preferred_ground_nodes).wait()

            new_state, _ = compact([component.state for component in self.components])

            return compacted_state - np.array(new_state)

        state_result = solve(problem, state0).x

        states = uncompact(state_result, state_compactification)
        for component, state in zip(self.components, states):
            component.state = state

        subsim = self.simulate(ts1, ts2, steps, preferred_ground_nodes)
        subsim.wait()

        with simulation as sim:
            sim.extend_from_simulation(subsim)

        simulation.finish()
        return simulation

    def multirate_steadystate(self,
                              t0: float,
                              *Ts: float,
                              fast_steps: int = 10,
                              subdivisions: Union[int, Collection] = 8,
                              periodicity_tolerance: float = 2**-26,
                              preferred_ground_nodes: Set = frozenset()) -> TransientSteadyStateEFMSimulation:
        """
        Finds the periodic steady-state solution of the circuit using the envelope-following-method (EFM) for
        multi-rate systems.

        The simulation is *asynchronous*! A future object is returned that allows you to access results and progress
        during simulation runtime. The same notes as for the standard `simulate` apply.

        Periodic steady-states occur for example for all linear circuits where complex analysis can be applied
        (i.e. circuits with sinusoidal sources, resistors, capacitors, inductors).
        This function requires you to pass ahead a suitable fundamental periods `T` (inverse of fundamental frequency).
        For multi-rate circuits (i.e. with many sinusoidal sources), `T` becomes the least common multiple of all
        those period times :math:`T` (the least common multiple is here extended to rational numbers - for non-rational
        frequencies, T can be approximated by rationalizing).

        In contrast to a single-rate system, multi-rate systems have multiple fundamental frequencies. If they are
        vastly different, simulation can be extremely time-consuming because the fastest frequency has to be simulated
        in accordance to not miss out on its effects.
        Usually, such systems' "main behavior" is only slightly influenced by the fast-rate frequency. For this case,
        the EFM can be utilized and many fast-rate intervals can be safely skipped to increase performance significantly
        without missing on accuracy.

        For the EFM-equations, see method `simulate_efm`.

        For reference, see the paper at https://ieeexplore.ieee.org/document/4342106.

        :param t0:
            The starting time.
        :param Ts:
            Fundamental period times (inverse of fundamental frequency) the circuit contains.
        :param fast_steps:
            The steps to simulate inside a single fast-rate interval.
        :param subdivisions:
            How many steps shall be simulated via EFM inside the slow-rate interval.
            Effectively determines the skip intervals for the EFM.

            * If of type `int`: Divides the fundamental period time equally.
            * If of type `typing.Collection`: Relative time points to place into the fundamental period time.
              Elements must be of type `float`.
        :param periodicity_tolerance
            A rounding tolerance to apply to the fundamental period times to obtain the overall
            covering fundamental frequency. The smaller this value, the rougher the simulation
            might estimate the whole steady state solution, but if too high insignificant floating
            point number precision may be caught up and lead to an extreme overall fundamental
            period time and again reduce simulation precision because for a longer fundamental
            frequency more subdivisions should be made over the interval to get a fairly correct
            solution.

            Especially reduce this value if you require lower period times than this tolerance, otherwise
            they will be effectively rounded to 0.
        :param preferred_ground_nodes:
            Notes to prefer as ground (GND).
        :return:
            A simulation object containing the steady-state solution.
            Note that computation happens *asynchronously*!
        """
        T_slow = flcm(*Ts, tol=periodicity_tolerance, ignore_zeros=True)
        T_fast = min(Ts)

        # Quantization step to find properly aligned intervals / skips that can be used for EF-method.
        if isinstance(subdivisions, int):
            if subdivisions < 0:
                raise ValueError('parameter subdivisions must be 0 or greater')

            discrete_subdivisions = np.linspace(0, T_slow // T_fast - 1, subdivisions + 2)
        elif isinstance(subdivisions, Collection):
            discrete_subdivisions = np.sort(np.fromiter(chain((s for s in subdivisions if 0.0 <= s <= 1.0), [0.0, 1.0]),
                                                        dtype=float)) * (T_slow // T_fast - 1)
        else:
            raise TypeError()

        interval_diffs = np.diff(np.floor(discrete_subdivisions).astype(int))
        skips = np.fromiter(compress(interval_diffs, interval_diffs > 0), dtype=int) - 1

        simulation = TransientSteadyStateEFMSimulation(skips, self._component_branches, self._graph.nodes)

        def multirate_steadystate_async(simulation, t0, T_fast, T_fast_steps, skips, preferred_ground_nodes):
            try:
                self._multirate_steadystate(simulation, t0, T_fast, T_fast_steps, skips, preferred_ground_nodes)
            except Exception as ex:
                simulation.abort(ex)

        Thread(target=multirate_steadystate_async,
               args=(simulation, t0, T_fast, fast_steps, skips, preferred_ground_nodes)).start()

        return simulation

    def _multirate_steadystate(self,
                               simulation: TransientSteadyStateEFMSimulation,
                               t0: float,
                               T_fast: float,
                               T_fast_steps: int,
                               skips: List[int],
                               preferred_ground_nodes: Set = frozenset()):

        state0, state_compactification = compact([component.state for component in self.components])

        def problem(first_interval_state):
            states = uncompact(first_interval_state, state_compactification)
            for component, state in zip(self.components, states):
                component.state = state

            self.simulate_efm(t0, T_fast, T_fast_steps, skips, preferred_ground_nodes).wait()

            final_interval_state, _ = compact([component.state for component in self.components])
            final_interval_state = np.array(final_interval_state)
            return first_interval_state - final_interval_state

        steadystate_vector = solve(problem, state0).x

        states = uncompact(steadystate_vector, state_compactification)
        for component, state in zip(self.components, states):
            component.state = state

        subsim = self.simulate_efm(t0, T_fast, T_fast_steps, skips, preferred_ground_nodes)
        subsim.wait()

        with simulation as sim:
            sim.extend_from_simulation(subsim)

        simulation.finish()
        return simulation



import numpy as np

from .TransientSimulation import TransientSimulation


class TransientSteadyStateSimulation(TransientSimulation):
    def get_steadystate(self, component) -> np.ndarray:
        """
        Returns the computed steady-state solution for a component.

        :param component:
            The component.
        :return:
            The steady-state vector.
        """
        return np.array(self.get_states(component)[0])



from threading import current_thread
from time import sleep, thread_time

import numpy as np


YIELD_TIMEOUT = np.nextafter(0, 1)


def yield_thread():
    sleep(YIELD_TIMEOUT)


def prefer_thread_yield(time: float = 0.01):
    """
    Requests to yield from current thread if runtime exceeds a certain amount of time.

    :param time:
        Time in seconds.
    """
    thread = current_thread()
    tm = thread_time()
    if hasattr(thread, '_last_threadtime_stamp'):
        if time < tm - thread._last_threadtime_stamp:
            thread._last_threadtime_stamp = tm
            yield_thread()
    else:
        thread._last_threadtime_stamp = tm



from dataclasses import dataclass

from .TwoTerminalCurrentComponent import TwoTerminalCurrentComponent


@dataclass(eq=False)
class Resistance(TwoTerminalCurrentComponent):
    """
    Describes an ideal resistance (R*I = V).

    value:
        The resistance value in Ohm.
    """
    value: float

    def get_current(self, v: float, t1: float, t2: float) -> float:
        return v / self.value

    def get_jacobian(self, v: float, t1: float, t2: float) -> float:
        return 1 / self.value


R = Resistance



from dataclasses import dataclass

from .TwoTerminalVoltageComponent import TwoTerminalVoltageComponent


@dataclass(eq=False)
class VoltageSourceDC(TwoTerminalVoltageComponent):
    """
    A DC voltage supply.

    Voltage is in the same direction as the branch points in the circuit graph.

    value:
        The voltage emitted by this supply (measured in Volt).
    """
    value: float

    def get_voltage(self, i: float, t1: float, t2: float) -> float:
        return self.value

    def get_jacobian(self, i: float, t1: float, t2: float) -> float:
        return 0



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Sequence

import numpy as np


@dataclass(eq=False)
class Branch:
    """
    The most basic element describing the relation between two potentials/terminals.

    Please use the more specific implementations `VoltageDrivenEdge` and `CurrentDrivenEdge` instead. Subclassing
    this base `Edge` class will have no meaning to the simulation algorithm that will simply ignore everything else.

    name:
        An optional name to set for the component. Can be set after initialization.
    """
    component: Component = field(repr=False)
    name: str = field(default=None, init=False)

    def __str__(self):
        branch_str = repr(self) if self.name is None or self.name == '' else self.name
        return f'{str(self.component)}/{branch_str}'


@dataclass(eq=False)
class CurrentBranch(Branch):
    def get_current(self, v_i: np.ndarray, t1: float, t2: float) -> float:
        """
        Calculated current depending on given voltages / currents and time steps.

        Shall return the current at present point in time `t2`.

        :param v_i:
            Voltages (measured in Volts) and currents (measured in Ampere). A vector filled with the potentials and
            branch currents as specified in the edges returned by `connect` (in order). See also `Element.connect`.
        :param t1:
            Previous point in time (measured in seconds).
        :param t2:
            Present point in time (measured in seconds).
        :return:
            The current (measured in Ampere).
        """
        raise NotImplementedError

    def get_jacobian(self, v_i: np.ndarray, t1: float, t2: float) -> Sequence[float]:
        """
        Returns the Jacobian matrix (i.e. the "gradient matrix") of the current function.

        For the numerical algorithm to quickly converge with more accuracy, it is necessary for electrical circuit
        problems to supply the derivations of each current / voltage function.

        This function specifically is supposed to return an array of derivatives instead of a matrix. Depending on the
        coupled branches specified by the holding component, the derivatives have to be returned in the same order.

        See also https://en.wikipedia.org/wiki/Jacobian_matrix_and_determinant

        Note that the Jacobian matrix must not include derivations over `t1` and `t2`.

        :param v_i:
            Voltages (measured in Volts) and currents (measured in Ampere). A vector filled with the potentials and
            branch currents as specified in the edges returned by `connect` (in order). See also `Element.connect`.
        :param t1:
            Previous point in time (measured in seconds).
        :param t2:
            Present point in time (measured in seconds).
        :return:
            The Jacobian (measured in Ampere per second).
        """
        raise NotImplementedError


@dataclass(eq=False)
class VoltageBranch(Branch):
    def get_voltage(self, v_i: np.ndarray, t1: float, t2: float) -> float:
        """
        Calculated current depending on given voltages / currents and time steps.

        Shall return the voltage at present point in time `t2`.

        :param v_i:
            Voltages (measured in Volts) and currents (measured in Ampere). A vector filled with the potentials and
            branch currents as specified in the edges returned by `connect` (in order). See also `Element.connect`.
        :param t1:
            Previous point in time (measured in seconds).
        :param t2:
            Present point in time (measured in seconds).
        :return:
            The current (measured in Ampere).
        """
        raise NotImplementedError

    def get_jacobian(self, v_i: np.ndarray, t1: float, t2: float) -> Sequence[float]:
        """
        Returns the Jacobian matrix (i.e. the "gradient matrix") of the voltage function.

        For the numerical algorithm to quickly converge with more accuracy, it is necessary for electrical circuit
        problems to supply the derivations of each current / voltage function.

        This function specifically is supposed to return an array of derivatives instead of a matrix. Depending on the
        coupled branches specified by the holding component, the derivatives have to be returned in the same order.

        See also https://en.wikipedia.org/wiki/Jacobian_matrix_and_determinant

        Note that the Jacobian matrix must not include derivations over `t1` and `t2`.

        :param v_i:
            Voltages (measured in Volts) and currents (measured in Ampere). A vector filled with the potentials and
            branch currents as specified in the edges returned by `connect` (in order). See also `Element.connect`.
        :param t1:
            Previous point in time (measured in seconds).
        :param t2:
            Present point in time (measured in seconds).
        :return:
            The Jacobian (measured in Volt per second).
        """
        raise NotImplementedError


@dataclass(eq=False)
class SwitchBranch(Branch):
    def switch_state(self, t: float) -> bool:
        """
        The current switch state at time `t`.

        :param t:
            Absolute point in time (measured in seconds).
        :return:
            `True` if switch is on, `False` if off.
        """
        raise NotImplementedError



import numpy as np
from scipy.optimize import check_grad

from .TaylorPolynomial import TaylorPolynomial


def linsys_jac(x, A, dA, db):
    r"""
    Computes the Jacobian matrix of a linear system whose :math:`A` and :math:`b` parts are functions.

    Consider a linear system with a random (vector) parameter :math:`y`:

    .. math::

        A(y) x = b(y)

    When the solution :math:`x` is known, the Jacobian :math:`J_x(y)` is then:

    .. math::

        J_x(y) = A^{-1} \left( \frac{\partial b}{\partial y} - \frac{\partial A}{\partial y} x \right)

    :param x:
        The solution of :math:`A(y) x = b(y)`.
    :param A:
        Linear system matrix :math:`A(y)`.
    :param dA:
        A 3rd-order tensor containing the element-wise derivatives of each matrix entry in :math:`A(y)`
        over each :math:`y`.
    :param db:
        The derivative matrix of :math:`b(y)`.
    :return:
    """
    return np.linalg.inv(A) @ (db - (dA @ x).transpose())


def check_jacobian(func, jac, x, *args, **kwargs):
    # Obtain dimension.
    fx = func(x, *args, **kwargs)

    def func_item(x, i, *args, **kwargs):
        return func(x, *args, **kwargs)[i]

    def jac_item(x, i, *args, **kwargs):
        return jac(x, *args, **kwargs)[i]

    return np.array([check_grad(func_item, jac_item, x, i, *args, **kwargs)
                     for i in range(len(fx))])



from dataclasses import dataclass
from math import exp

from .TwoTerminalCurrentComponent import TwoTerminalCurrentComponent


@dataclass(eq=False)
class ShockleyDiode(TwoTerminalCurrentComponent):
    """
    Describes an ideal Schockley Diode backed by the ideal Schockley equation.

    See https://en.wikipedia.org/wiki/Diode_modelling#Shockley_diode_model

    For a simpler, linearized model with better solution convergence properties,
    see `respice.components.LinearizedShockleyDiode`.

    i_s:
        Reverse bias saturation current (also known as leakage current).
        Depending on the material used, common values are
        * silicon diodes: usually between 10^-12 A and 10^-10 A (1uA to 100uA).
        * germanium diodes: about 10^-4 A (0.1mA)
    v_t:
        Thermal voltage V_T. For approximate room temperate 300K this value
        is 25.85mV (0.02585V). This is default.
    n:
        Ideality factor (also known as quality factor). By default 1.
    """
    i_s: float = 1e-12
    v_t: float = 0.02585
    n: float = 1.0

    def get_current(self, v: float, t1: float, t2: float) -> float:
        return self.i_s * (exp(v / (self.n * self.v_t)) - 1)

    def get_jacobian(self, v: float, t1: float, t2: float) -> float:
        return self.i_s / (self.n * self.v_t) * exp(v / (self.n * self.v_t))



from bisect import bisect_left
from math import pi
from types import SimpleNamespace
from typing import Optional, Iterable

import numpy as np

from .PeriodicVoltageSource import PeriodicVoltageSource


class VoltageSourcePWM(PeriodicVoltageSource):
    """
    A periodic, multi-level PWM modulation source.

    You can supply this source a constant pattern that is repeated throughout a simulation.

    Changing patterns on the same source object is currently not possible. Please instantiate another object with
    the new pattern.
    """
    def __init__(self, pattern_levels: Iterable[float], pattern_timings: Iterable[float], *args, **kwargs):
        r"""
        :param pattern_levels:
            The amplitude values of the pattern.
        :param pattern_timings:
            The timings between each amplitude change in `pattern_levels` in rad. The range for pattern timings lies
            inside :math:`(0, 2\pi)`. Timings outside this range will be ignored.
            Note that the length of this iterable must be always one less than `len(pattern_levels)`.
            This parameter *must* be sorted beforehand!
        """
        super().__init__(*args, **kwargs)

        self._pattern_timings = list(pattern_timings)
        self._pattern_levels = list(pattern_levels)

        assert len(self._pattern_timings) + 1 == len(self._pattern_levels)

        self._cache = SimpleNamespace()
        self._cache.repr_pattern = None

    def __repr__(self):
        if self._cache.repr_pattern is None:
            # Generate human readable pattern for repr().
            pattern_length = 50
            minv = min(self._pattern_levels)
            maxv = max(self._pattern_levels)
            chardict = {
                0: '\u2581',
                1: '\u2582',
                2: '\u2583',
                3: '\u2584',
                4: '\u2585',
                5: '\u2586',
                6: '\u2587',
                7: '\u2588',
                8: '\u2588',
            }

            self._cache.repr_pattern = ''.join(
                chardict[int((self.get_signal(phase) - minv) / (maxv - minv) * (len(chardict) - 1))]
                for phase in np.linspace(0, 2 * pi, pattern_length))

        return f'{type(self).__name__}(pattern: {self._cache.repr_pattern})'

    def get_signal(self, phase: float) -> float:
        return self._pattern_levels[bisect_left(self._pattern_timings, phase)]

    def next_signal_event(self, phase1: float, phase2: float) -> Optional[float]:
        if phase1 == 0.0 and self._pattern_levels[0] != self._pattern_levels[-1]:
            return 0.0

        a = bisect_left(self._pattern_timings, phase1)
        b = bisect_left(self._pattern_timings, phase2, a)

        if a == b:
            return None

        return self._pattern_timings[a]



from dataclasses import dataclass
from math import pi
from typing import Optional, Callable

import numpy as np

from .TwoTerminalVoltageComponent import TwoTerminalVoltageComponent


@dataclass(eq=False)
class PeriodicVoltageSource(TwoTerminalVoltageComponent):
    """
    A periodic voltage supply.

    This class is a building block to define your own voltage/current invariant periodic sources easily. Subclasses
    override the `get_signal()` function that supplies the signal values inside a single period.

    frequency:
        The current frequency (measured in Hertz).
    phase:
        The initial phase angle (measured in rad).
    """
    frequency: float = 1.0
    phase: float = 0.0

    @property
    def T(self):
        """
        The period time :math:`T`.
        """
        return 1.0 / self.frequency

    @T.setter
    def T(self, value):
        self.frequency = 1.0 / value

    def get_phase_from_time(self, t):
        return t * 2 * pi * self.frequency + self.phase

    def get_voltage(self, i: float, t1: float, t2: float) -> float:
        return self.get_signal(self.get_phase_from_time(t2) % (2 * pi))

    def get_jacobian(self, i: float, t1: float, t2: float) -> float:
        return 0

    def get_signal(self, phase: float) -> float:
        """
        Returns the signal.

        :param phase:
            The momentary angular phase, inside the range [0, 2pi).
        :return:
            The signal value at `phase`.
        """
        raise NotImplementedError

    def next_event(self, i: Callable[[float], np.ndarray], t1: float, t2: float) -> Optional[float]:
        interval1, phase1 = divmod(self.get_phase_from_time(t1), 2 * pi)
        interval2, phase2 = divmod(self.get_phase_from_time(t2), 2 * pi)

        intv_phase1 = phase1
        intv_phase2 = np.nextafter(2 * pi, 0)
        for p in range(int(interval1), int(interval2)):
            eventphase = self.next_signal_event(intv_phase1, intv_phase2)
            if eventphase is not None and 0 <= eventphase < 2 * pi:
                return (p + (eventphase - self.phase) / (2 * pi)) / self.frequency

            intv_phase1 = 0

        eventphase = self.next_signal_event(intv_phase1, phase2)
        return None if eventphase is None else (int(interval2) + (eventphase - self.phase) / (2 * pi)) / self.frequency

    def next_signal_event(self, phase1: float, phase2: float) -> Optional[float]:
        r"""
        If a source is discontinuous, this function should be used to indicate such discontinuities
        as events.

        Instead of event times, phases indicating the event are returned.
        They are properly translated to according simulation times.

        This function guarantees to serve phases in the range :math:`[0, 2\pi)`.

        :param phase1:
            Previous phase (measured in rad).
        :param phase2:
            Present phase (measured in rad).
        :return:
            The next phase at which an event occurs, or None if no event happens.
        """
        return None



from __future__ import annotations

from dataclasses import dataclass, field
from typing import List, Hashable, Tuple, Optional, Callable

import numpy as np

from .Branch import Branch


@dataclass(eq=False)
class Component:
    """
    The base class for all electrical components.

    Each component consists of at least one `Edge`. An edge represents a potential difference between two terminals.
    Since a component can be multi-terminal, it must itself specify what edges are created between a given amount
    of terminal by the user. Those terminals are passed in order to the `connect` function. See `connect()`.

    name:
        An optional name to set for the component. Can be set after initialization.
    """
    name: str = field(default=None, init=False)

    def connect(self, *terminals: Hashable) -> List[Tuple[Hashable, Hashable, Branch]]:
        """
        Specifies how the component gets connected in the underlying circuit graph representation.

        When calling `circuit.add_element`, the order of the terminals specified matters. The elements `connect`
        function is called with the terminals to connect with in order. This function shall establish edges in the
        graph representation of the circuit and add so called "couplings". Couplings are equations that further
        constraint the behaviour between different edges (or the behaviour on the edge itself, then the element
        is "self-coupled").

        For example:
        * A resistance is a self-coupled element. Its current is immediately depending on its voltage.
        * A coupled inductance is a mutually-coupled element. Its current depends on the voltage of the other side
        and vice versa. The edges describing the two coils on each side are mutually coupled by one
        additional equation each.

        :param graph:
            The graph to establish edges/connections in.
        :param terminals:
            The potentials (aka terminals) to connect the element's ports with.
        :return:
            A list of edges in the form of `(terminal1, terminal2, edge)`. The order is respected in the voltage-vector
            when invoking each edge's `get_current` function.
        """
        raise NotImplementedError("I don't know how to connect this element between the given terminals")

    def update(self, v_i: np.ndarray, t1: float, t2: float):
        """
        Updates the state of the component.

        This method is called after each simulation cycle in the `Circuit.simulate` function.
        If the simulation algorithm was able to determine a proper voltage and current,
        this method is called to set those calculated values as the new "truth" for the
        next time step.

        :param v_i:
            The voltages and currents of all other coupled edges (measured in Volts and Amperes respectively).
        :param t1:
            Previous point in time (measured in seconds).
        :param t2:
            Present point in time (measured in seconds).
        """
        pass

    def next_event(self, vi: Callable[[float], np.ndarray], t1: float, t2: float) -> Optional[float]:
        r"""
        Calculates a possible next event happening in between a calculation step.

        An event is a significant point in time to consider to stop simulation at and perform an
        additional step computation. Such significant events are for example discontinuities.

        This function let's the user handle solving for events inside their respective components.
        Due to different natures and behaviors of components, the solution can be better tied
        individually (e.g. for continuous problems, how interpolation over step solution is performed
        or immediate calculation for periodic sources).

        :param vi:
            The input voltages and currents as fed to the inner branches.

            The parameter is a callable representing an interpolation function between `t1` and `t2`.
            Since between `t1` and `t2` the exact solution is not known, the solver provides a suitable interpolation
            function that can be used to calculate an event continuously, because often events also depend on the state
            of a component.

            It is not recommended to use values beyond the range :math:`[\texttt{t1}, \texttt{t2}]` because they are
            extrapolated outside and may strongly deviate from the actual results.
        :param t1:
            Previous point in time (measured in seconds).
        :param t2:
            Present point in time (measured in seconds).
        :return:
            A time :math:`t` between `t1` and `t2` or `None` when no event would occur.

            Broken events that lie outside the range :math:`[\texttt{t1}, \texttt{t2}]` will be ignored.
        """
        return None

    @property
    def state(self):
        """
        The whole state of the component.

        A few components (such as inductors or capacitors) have internal state, i.e. depending on the inner state,
        calculated currents and voltages change.

        This function provides a general interface to access the state of components. It is especially used to calculate
        steady-state solutions where the solver must know the state of the components to determine whether indeed a
        steady state has been found.

        If your component does not have any state, do not override this property.

        :return:
            A list or tuple of the state values. Be sure that the order of the state variables does not change.

            When the component has no state, an empty tuple is returned by default.
        """
        return tuple()

    @state.setter
    def state(self, val):
        """
        Sets the component's state.

        :param val:
            The new state.
        """
        pass

    @property
    def default_branch(self):
        """
        Returns the default or main branch of the component. If `None` no default branch is set.

        This field is used for convenient lookup of simulation results by component.
        """
        return None

    def __str__(self):
        if self.name is None or self.name is '':
            return repr(self)
        return self.name



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Hashable, List, Tuple, Sequence

import numpy as np

from .Branch import Branch, CurrentBranch, VoltageBranch
from .Component import Component


@dataclass(eq=False)
class OperationalAmplifier(Component):
    """
    Describes an (almost) ideal operational amplifier ("op-amp").

    An ideal op-amp has following idealizing characteristics:
    * Infinite open-loop gain (this is the only aspect which does not hold for this op-amp implementation, it's finite)
    * Infinite input impedance, and so zero input current
    * Zero input offset voltage
    * Infinite output voltage range
    * Infinite bandwidth with zero phase shift and infinite slew rate
    * Zero output impedance
    * Zero noise

    See also https://en.wikipedia.org/wiki/Operational_amplifier#Ideal_op-amps

    This component is connected in following order:
    * First two terminals denote the input voltages :math:`V_{in-}` and :math:`V_{in+}`.
    The difference between them is what is amplified.
    * The third terminal denotes GND (ground) or any reference potential you want to use.
    * The fourth terminal denotes the output voltage V_out (in reference to GND).

    g:
        Open loop gain. For real op-amps this value lies between 20000 and 200000.
    """

    @dataclass(eq=False)
    class InputBranch(CurrentBranch):
        def get_current(self, v_i: np.ndarray, t1: float, t2: float) -> float:
            return 0

        def get_jacobian(self, v_i: np.ndarray, t1: float, t2: float) -> Sequence[float]:
            return [0, 0]

    @dataclass(eq=False)
    class OutputBranch(VoltageBranch):
        def get_voltage(self, v_i: np.ndarray, t1: float, t2: float) -> float:
            return self.component.g * v_i[0]

        def get_jacobian(self, v_i: np.ndarray, t1: float, t2: float) -> Sequence[float]:
            return [self.component.g, 0]

    g: float = 100000
    _input_branch: InputBranch = field(init=False, repr=False)
    _output_branch: OutputBranch = field(init=False, repr=False)

    def __post_init__(self):
        self._input_branch = self.InputBranch(self)
        self._input_branch.name = 'input'
        self._output_branch = self.OutputBranch(self)
        self._output_branch.name = 'output'

    def connect(self,
                v_in_neg: Hashable,
                v_in_pos: Hashable,
                gnd: Hashable,
                v_out: Hashable) -> List[Tuple[Hashable, Hashable, Branch]]:
        """
        Connects the transistor.

        :param v_in_neg:
            The negative terminal of the input voltage to amplify.
        :param v_in_pos:
            The positive terminal of the input voltage to amplify.
        :param gnd:
            The ground / reference potential.
        :param v_out:
            The output voltage (in reference to given GND)
        """
        return [
            (v_in_neg, v_in_pos, self._input_branch),
            (gnd, v_out, self._output_branch),
        ]

    @property
    def input(self):
        return self._input_branch

    @property
    def output(self):
        return self._output_branch



import numpy as np
from scipy.optimize import minimize


def solve(fun, x0, jac=None, args=tuple(), method='Nelder-Mead'):
    def f(x, *args):
        fx = fun(x, *args)

        if jac is None:
            return np.sum(np.square(fx))

        if jac is True:
            fx, j = fx
        else:
            j = jac(fx, *args)

        resid = np.sum(np.square(fx))
        resid_grad = 2 * (np.transpose(j) @ fx)

        return resid, resid_grad

    return minimize(f, x0, args=args, jac=None if jac is None else True, method=method)



from fractions import Fraction
from functools import reduce

import numpy as np


def lcm(*numbers: int):
    """
    Returns the least common multiple of all given numbers.
    """
    return reduce(lambda x, y: np.lcm(x, y), numbers)


def flcm(*numbers: float, tol=2**-26, ignore_zeros=False):
    r"""
    Returns the least common multiple for floating point numbers.

    The least common multiple for floating point numbers is not exactly analogous to the
    normal case. Mainly, if rational numbers are considered, pure rational numbers would
    always produce :math:`\infty`. A viable solution only exists for rational numbers.

    In fact, computer floating point precision is always a rational numbers. However, due to
    rounding and other effects, it is usually desired to obtain an approximate value for the
    least common multiple. This can be controlled by the parameter `tol`.

    :param numbers:
        The floating point numbers to get the least common multiple from.
    :param tol:
        The tolerance. The floating point numbers are converted to `Fraction`s and rounded
        to the closes fraction fulfilling this tolerance.
    :param ignore_zeros:
        Whether to ignore zero values encountered in `numbers`.
    """
    fractional_limit = int(1 / tol)
    fractions = [Fraction(number).limit_denominator(fractional_limit)
                 for number in numbers]

    if ignore_zeros:
        fractions = [fraction for fraction in fractions if fraction.numerator != 0]

    numerators = np.fromiter((f.numerator for f in fractions), dtype=int)
    denominators = np.fromiter((f.denominator for f in fractions), dtype=int)

    common_denominator = lcm(*denominators)

    normalized_numerators = numerators * (common_denominator // denominators)

    return lcm(*normalized_numerators) / common_denominator



from respice.analysis import Circuit
from respice.components import Capacitance, Inductance


class LC(Circuit):
    """
    Example ideal LC circuit.
    """

    def __init__(self, l: float, c: float, initial_v: float = 1.0, initial_i: float = 0.0):
        """
        :param l:
            The inductance value (measured in Henry).
        :param c:
            The capacitance value (measured in Farad).
        :param initial_v:
            The initial voltage of the capacitance (measured in Volts).
        :param initial_i:
            The initial current of the inductance (measured in Amperes).
        """
        super().__init__()

        self.L = Inductance(l, initial_i)
        self.L.name = 'L'
        self.C = Capacitance(c, initial_v)
        self.C.name = 'C'

        self.add(self.L, 0, 1)
        self.add(self.C, 1, 0)
